from figcli.models.sso.okta.okta_auth import OktaAuth
from figcli.models.sso.okta.okta_session import OktaSession


class OktaSessionAuth(OktaAuth):
    def __init__(self, session: OktaSession):
        self.session = session

    def get_session(self):
        return self.session




class Role:

    def __init__(self, role: str, full_name: str = None):
        self.role: str = role
        self.full_name = full_name

    def __str__(self):
        return self.role

    def __eq__(self, other):
        return self.role == other.role

    def __hash__(self):
        return hash(self.role)



import re
from typing import Set, Dict
from figcli.config import *


class KeyUtils(object):

    @staticmethod
    def find_all_expected_names(config_keys: set, shared_names: set, merge_conf: dict,
                                repl_conf: dict, repl_from_conf: dict, namespace: str) -> Set[str]:
        """
        From various sets of keys and configs, calculates all required PS Names (keys) this application requires.
        Args:
            config_keys: set -> representing the app_figs in the passed in figgy.json file
            shared_names: set -> representing the shared_figs in the passed in figgy.json file
            merge_conf: dict -> representing the merged_figs in the passed in figgy.json file
            repl_conf: dict -> representing the replicate_figs in the figgy.json file
            repl_from_conf: dict -> represents the `replicate_from` config block in the figgy.json file
            namespace: parsed, or calculated namespace for the application being synced. E.G. /app/demo-time/

        Returns: Set[str] -> All PS Names that have been defined as dependencies for this application's deployment
        """
        merge_keys = set(merge_conf.keys())
        all_keys = config_keys | shared_names | merge_keys

        for merge_key, merge_val in merge_conf.items():
            if type(merge_val) == list:
                for val in merge_val:
                    if KeyUtils.sanitize(val).startswith(namespace):
                        all_keys.add(KeyUtils.sanitize(val))
            elif type(merge_val) == str:
                matches = re.findall('\${([\w/-]+)}', merge_val)
                for match in matches:
                    if match.startswith(namespace):
                        all_keys.add(KeyUtils.sanitize(match))

        for key in repl_conf:
            all_keys.add(repl_conf[key])

        source_ns = repl_from_conf.get(SOURCE_NS_KEY)
        params = repl_from_conf.get(PARAMETERS_KEY)

        if source_ns and params:
            for param in params:
                all_keys.add(f'{namespace}{param}')

        return all_keys


    @staticmethod
    def merge_repl_and_repl_from_blocks(repl_conf: Dict, repl_from: Dict, dest_namespace: str) -> Dict:
        """
        Parses the repl_from block and merges it into the standard 'replication' block. This simplifies
        configuring replication and detecting orphans.
        Args:
            repl_conf: Dict representing the `replicate_figs` block in figgy.json file
            repl_from: Dict representing the `replicate_frorm` block in figgy.json file
            dest_namespace: namespace found in the `figgy.json` file. Where replication is destined for.

        Returns: an updated repl_conf dictionary with repl_from figcli.configs merged into it.
        """

        source_ns = repl_from.get(SOURCE_NS_KEY)
        params = repl_from.get(PARAMETERS_KEY)
        dest_namespace = dest_namespace if dest_namespace.endswith('/') else f'{dest_namespace}/'

        if source_ns and params:
            source_ns = source_ns if source_ns.endswith('/') else f'{source_ns}/'
            for param in params:
                repl_conf[f'{source_ns}{param}'] = f'{dest_namespace}{param}'

        return repl_conf

    @staticmethod
    def sanitize(merge_val):
        return merge_val.replace("${", "").replace("}", "").replace(merge_uri_suffix, "")

    @staticmethod
    def desanitize(merge_val):
        return "${" + merge_val + "}"



from figcli.config import *


class Account:

    def __init__(self, account: str):
        self.account = account

    def __str__(self):
        return self.account

    def __eq__(self, other):
        return self.account == other.account



import sys

import pexpect

from figcli.test.cli.config import *
from figcli.test.cli.data.delete import DataDelete, DevGet
from figcli.test.cli.figgy import FiggyTest
from figcli.utils.utils import *


class DataSync(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(None, extra_args=extra_args)

    def run(self):
        self.step("Prepping workspace, deleting any existing values.")
        self.prep()

        child = pexpect.spawn(f'{CLI_NAME} {Utils.get_first(config)} {Utils.get_first(sync)} '
                              f'--env {DEFAULT_ENV} --config figcli/test/assets/data_repl_conf.json '
                              f'--skip-upgrade --replication-only {self.extra_args}', timeout=10, encoding='utf-8')
        time.sleep(2)
        with open('figcli/test/assets/data_repl_conf.json', 'r') as file:
            content = json.loads(file.read()).get('replicate_figs')

        self.step("Running replication-only sync and adding values.")
        sources = list(content.keys())
        for source in sources:
            child.expect(f".*input a value.*")
            child.sendline(DELETE_ME_VALUE)
            child.expect(".*description.*")
            child.sendline("desc")
            child.expect(".*secret.*")
            child.sendline('y')
            child.expect('.*encryption.*')
            child.sendline('data')
            child.expect('.*added successfully.*')

        self.step("Waiting for replication propagation")
        time.sleep(15)

        self.step("Verifying successful replication")
        destinations = list(content.values())
        for dest in destinations:
            DevGet().get(dest, DELETE_ME_VALUE, get_more=False, no_decrypt=True)

        self.step("Cleaning leftover values")
        self.prep()

    def prep(self):
        with open('figcli/test/assets/data_repl_conf.json', 'r') as file:
            content = json.loads(file.read()).get('replicate_figs')

        delete = DataDelete()
        destinations = list(content.values())
        for i in range(0, len(destinations)):
            print(f"DELETING: {destinations[i]}")

            if i < len(destinations):
                delete.delete(destinations[i], delete_another=True, repl_dest_delete=True)
            else:
                delete.delete(destinations[i], delete_another=False, repl_dest_delete=True)

        sources = list(content.keys())
        print(f"SOURCES: {sources}")
        for i in range(0, len(sources)):
            print(f"DELETING: {destinations[i]}")
            if i < len(sources):
                delete.delete(sources[i], delete_another=True)
            else:
                delete.delete(sources[i], delete_another=False)



from collections import namedtuple
from typing import Optional

from boto3.session import Session

from figcli.commands.iam_context import IAMContext
from figcli.commands.types.iam import IAMCommand
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.utils.utils import *
from figcli.utils.awscli import AWSCLIUtils


class Export(IAMCommand):
    """
    Returns audit history for a queried PS Name
    """

    def __init__(self, iam_context: IAMContext, env_session: Session,
                 all_sessions: Optional[Dict[str, Session]]):
        super().__init__(export, iam_context)
        self._all_sessions: Optional[Dict[str, Session]] = all_sessions
        self._env_session: Session = env_session

    def _export(self):
        if not self._all_sessions:
            credentials: namedtuple = self._env_session.get_credentials().get_frozen_credentials()
            Utils.stc_validate(credentials is not None,
                               f"Unable to generate credentials for environment: {self.run_env}")
            AWSCLIUtils.write_credentials(credentials.access_key, credentials.secret_key, credentials.token,
                                          region=self.context.defaults.region, color=self.c)
        else:
            for (role_name, session) in self._all_sessions.items():
                credentials: namedtuple = session.get_credentials().get_frozen_credentials()
                Utils.stc_validate(credentials is not None, f"Unable to generate credentials for role: {role_name}")
                AWSCLIUtils.write_credentials(credentials.access_key, credentials.secret_key, credentials.token,
                                              profile_name=role_name, region=self.context.defaults.region, color=self.c)

    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._export()



from figcli.config.commands import *
from figcli.config.constants import *
from figcli.config.figgy import *
from figcli.utils.collection_utils import CollectionUtils

# User communication text
CLEANUP_REPLICA_ORPHANS = "Some orphaned replication mappings were discovered for your service. " \
                          "Once they are no longer necessary please clean them up using the `cleanup` command."
SHARED_NAME_RESOLUTION_MESSAGE = f"Your application is missing at least one ParameterStore name that it expects to " \
    f"exist as defined in either the {SHARED_KEY} block of your configuration, or as a " \
    f"dependency of a merge key. To resolve " \
    f"this, contact whoever is responsible for sharing these config value(s) with you " \
    f"and have them setup up config replication of the values into your apps namespace " \
    f"at the expected path."
RESOURCE_PARSER_DESC = "Provides utilities to help manage application configs and secrets across all environments. " \
    f" For more details run `{CLI_NAME} [resource] --help"
COMMAND_PARSER_DESC = "Provides utilities to help manage application configs and secrets across all environments. " \
    f"For more details on commands, run: `{CLI_NAME} [resource] [command] --help`"
CONFIG_REQ_TEXT = f"--config is a requirement argument."
CI_CONFIG_HELP_TEXT = f"Path to your project's figgy.json file."
UNUSED_CONFIG_DETECTED = f"%%red%%The following Names were found in PS but are not referenced in your configurations. \n" \
    f"Use the %%rs%%%%blue%%`cleanup`%%rs%%%%red%% command to clean them up once all " \
    f"deployed application versions no longer use these configurations: %%rs%%"

# Help Text
VERSION_HELP_TEXT = f'Prints current version, which is, in this case: {VERSION}'
UPGRADE_HELP_TEXT = f'If available for your operating system, walks the user through an automatic upgrade process.'
SKIP_UPGRADE_HELP_TEXT = f'Prevents the figgy for checking for a new version. Useful when running E2E tests locally.'
COMMAND_HELP_TEXT = f'Valid values are: {CollectionUtils.printable_set(config_commands)}'
RESOURCE_HELP_TEXT = f'Valid values are: {CollectionUtils.printable_set(resources)}'
CONFIG_HELP_TEXT = f"Managed configuration resources. Commands: [{CollectionUtils.printable_set(config_commands)}]"
PROMPT_HELP_TEXT = f"With --prompt set you will always be prompted for your AWS CLI Profile name and user type."
REPLICATION_ONLY_HELP_TEXT = f"Sync a declarative replication-config json file that _only_ contains the {REPLICATION_KEY} block."
MANUAL_HELP_TEXT = f"Migrate any K/V Hierarchy, or a single KV Pair from Consul into anywhere in ParameterStore " \
    f"(that you have permissions)."
IAM_HELP_TEXT = f"Manage your temporary credentials. [{CollectionUtils.printable_set(iam_commands)}]"
ENV_HELP_TEXT = f"Valid values are:"
INFO_HELP_TEXT = f"Prints out more detailed information about the selected subcommand."
CONFIGURE_HELP_TEXT = f"(Optional) Setup default configurations to reduce redundant prompts."
EMPTY_ENV_HELP_TEXT = f"No --env parameter found. Using default: "
DELETE_HELP_TEXT = f"Manually delete parameters from Parameter Store"
LIST_HELP_TEXT = f"Lists all matching configuration Names based on a namespace prefix. e.g. /app/demo-time/"
PUT_HELP_TEXT = f"Store (or update) an arbitrary configuration in ParameterStore."
SHARE_HELP_TEXT = f"Share a parameter from a source to an arbitrary destination path in Parameter Store via config " \
    f"replication."
CLEANUP_HELP_TEXT = 'Cleanup will compare your config desired state as defined in your figgy.json with the current ' \
                    'config state in AWS. You will be prompted on whether or not you wish to delete orphaned ' \
                    'configurations (configs that exist in AWS but not in your figgy.json file).'
MIGRATE_HELP_TEXT = f"Walks a user through migrating application configs from Consul into Parameter Store"
SYNC_HELP_TEXT = "Synchronizes your defined figgy.json configurations with those in Parameter Store. " \
                 "This ensures the proper configurations exist in the provided run environment for your application."
GET_HELP_TEXT = "Retrieve an arbitrary value from ParameterStore by Name."
BROWSE_HELP_TEXT = "Browse and look up or delete parameters through a tree structure."
AUDIT_HELP_TEXT = "Audit parameter store changes to parameters."
DUMP_HELP_TEXT = "Dump a series of PS Values by a queried prefix as JSON, or output to a file."
PREFIX_HELP_TEXT = "The prefix (e.g. /app/demo-time) to limit results to."
OUT_HELP_TEXT = "File to write the outputted data to. e.g.: --out /tmp/some-file.json"
EXPORT_HELP_TEXT = "Writes temporary STS AWS credentials to your ~/.aws/credentials file under the [default] profile."

PROMOTE_HELP_TEXT = "Promote a set of configurations under an arbitrary app namespace to a higher environment."
RESTORE_HELP_TEXT = "Restore a single parameter or all parameters within a time range"
DEBUG_HELP_TEXT = "Turn on debug mode for enhanced logging."
EDIT_HELP_TEXT = "Edit the value of an existing parameter store parameter, or create a new one"
COPY_FROM_HELP_TEXT = "Copy values from another Parameter Store namespace."
GENERATE_HELP_TEXT = "Generate a new figgy.json file from an existing figgy.json file. This is useful when you " \
                     "want to deploy a single service multiple times under different names with slightly differing " \
                     "configurations."
FROM_HELP_TEXT = "/path/to/figgy.json file to use as the source template file to generate from. Your generated file" \
                 " will automatically have all of the configurations from the source-file configured to replicate to it."
LOGIN_HELP_TEXT = "Logs you in to all viable AWS accounts and caches session locally. This will prevent you from " \
                  "having to input your MFA every time you swap into a new account or role. \n\nAlternate Use:" \
                  "`figgy login sandbox` will log you into the Figgy sandbox playground for hassle-free " \
                  "experimentation."
VALIDATE_HELP_TEXT = "Validates a `figgy.json` file and ensures all required configurations exist in ParameterStore. " \
           "Exits with an error code if any defined configuration is missing."
LOGIN_SANDBOX_HELP_TEXT = "Get a temporary session from the free figgy sandbox and have fun, experiment, do whatever" \
                          "you want!"

MISSING_PS_NAME_MESSAGE = "Your application is missing at least one ParameterStore name that it expects to exist as " \
                          "defined in your figgy.json file. To resolve this issue, first try running your `figgy` " \
                          " `sync` command to validate and synchronize your expected configs."
PROFILE_HELP_TEXT = "Overrides all other figgy configurations and instead pulls ALL credentials from your local " \
                    "~/.aws/credentials file and uses that for all types of authorization. Ideal for CICD pipelines."


# Point in time (--point-in-time)
POINT_IN_TIME = "Restore all parameters to a point in time."

ALL_PROFILES = "Export all available AWS profiles to ~/.aws/credentials"
ROLE = "Specify role to run command with"

HELP_TEXT_MAP = {
    version: VERSION_HELP_TEXT,
    command: COMMAND_HELP_TEXT,
    resource: RESOURCE_HELP_TEXT,
    config: CONFIG_HELP_TEXT,
    prompt_com: PROMPT_HELP_TEXT,
    replication_only: REPLICATION_ONLY_HELP_TEXT,
    manual: MANUAL_HELP_TEXT,
    env: ENV_HELP_TEXT,
    info: INFO_HELP_TEXT,
    configure: CONFIGURE_HELP_TEXT,
    delete: DELETE_HELP_TEXT,
    list_com: LIST_HELP_TEXT,
    put: PUT_HELP_TEXT,
    share: SHARE_HELP_TEXT,
    cleanup: CLEANUP_HELP_TEXT,
    sync: SYNC_HELP_TEXT,
    get: GET_HELP_TEXT,
    browse: BROWSE_HELP_TEXT,
    audit: AUDIT_HELP_TEXT,
    dump: DUMP_HELP_TEXT,
    out: OUT_HELP_TEXT,
    prefix: PREFIX_HELP_TEXT,
    skip_upgrade: SKIP_UPGRADE_HELP_TEXT,
    restore: RESTORE_HELP_TEXT,
    point_in_time: POINT_IN_TIME,
    export: EXPORT_HELP_TEXT,
    iam: IAM_HELP_TEXT,
    promote: PROMOTE_HELP_TEXT,
    all_profiles: ALL_PROFILES,
    role: ROLE,
    debug: DEBUG_HELP_TEXT,
    edit: EDIT_HELP_TEXT,
    copy_from: COPY_FROM_HELP_TEXT,
    generate: GENERATE_HELP_TEXT,
    from_path: FROM_HELP_TEXT,
    login: LOGIN_HELP_TEXT,
    sandbox: LOGIN_SANDBOX_HELP_TEXT,
    validate: VALIDATE_HELP_TEXT,
    profile: PROFILE_HELP_TEXT,
    upgrade: UPGRADE_HELP_TEXT
}

# Other
SNS_EMAIL_SUBJECT = "Figgy error detected."
DESC_MISSING_TEXT = "None specified"

# Prompts:
is_secret = [
    ('class:', 'Is this value a secret? (y/N): ')
]


from abc import ABC

from figcli.commands.help_context import HelpContext
from figcli.commands.types.command import Command


class HelpCommand(Command, ABC):
    """
    Help command class to support some commands that may not have a resource or specific command selected
    """

    def __init__(self, command_type: frozenset, colors_enabled: bool, context: HelpContext):
        super().__init__(command_type, colors_enabled, context)



from figcli.utils.utils import *
import platform
import keyring
import pyotp
from figcli.config import *
from keyring.backends.OS_X import Keyring
from keyring.backends.Windows import WinVaultKeyring
from keyrings.alt.file import EncryptedKeyring, PlaintextKeyring
import os


class SecretsManager:

    @staticmethod
    def set_keyring():
        if platform.system() == WINDOWS:
            keyring.set_keyring(WinVaultKeyring())
        elif os.environ.get(OVERRIDE_KEYRING_ENV_VAR) == "true":  # Used in builds when running tests
            keyring.set_keyring(PlaintextKeyring())
        elif platform.system() == MAC:
            keyring.set_keyring(keyring.backends.OS_X.Keyring())
        elif platform.system() == LINUX:
            keyring.set_keyring(EncryptedKeyring())
        else:
            Utils.stc_error_exit("Only OSX and MAC and Linux with installed SecretStorage are supported for "
                                 "OKTA + Keyring integration.")

    @staticmethod
    def generate_mfa(user: str) -> str:
        SecretsManager.set_keyring()
        mfa_secret = keyring.get_password(FIGGY_KEYRING_NAMESPACE, f'{user}-mfa')
        token = pyotp.TOTP(mfa_secret).now()
        print(f"Authenticating with one-time token: {token}")
        return token

    @staticmethod
    def set_mfa_secret(user: str, mfa_secret: str):
        SecretsManager.set_keyring()
        keyring.set_password(FIGGY_KEYRING_NAMESPACE, f'{user}-mfa', mfa_secret)

    @staticmethod
    def set_password(user: str, password: str) -> None:
        SecretsManager.set_keyring()
        keyring.set_password(FIGGY_KEYRING_NAMESPACE, user, password)

    @staticmethod
    def get_or_set(user: str, backup: str):
        current = SecretsManager.get_password(user)
        if not current:
            SecretsManager.set_password(user, backup)

        return current

    @staticmethod
    def get_password(user: str) -> str:
        if platform.system() == WINDOWS:
            keyring.set_keyring(WinVaultKeyring())
        elif os.environ.get(OVERRIDE_KEYRING_ENV_VAR) == "true":  # Used in circleci builds for running tests
            keyring.set_keyring(PlaintextKeyring())
        elif platform.system() == MAC:
            keyring.set_keyring(keyring.backends.OS_X.Keyring())
        elif platform.system() == LINUX:
            keyring.set_keyring(EncryptedKeyring())
        else:
            Utils.stc_error_exit("Only OSX and MAC are supported for OKTA + Keyring integration.")

        return keyring.get_password(FIGGY_KEYRING_NAMESPACE, user)



from figcli.commands.help_context import HelpContext
from figcli.commands.types.help import HelpCommand
from figcli.config import *
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker


class Version(HelpCommand):
    """
    Drives the --version command
    """

    def __init__(self, help_context: HelpContext):
        super().__init__(version, help_context.defaults.colors_enabled, help_context)
        self.tracker = VersionTracker(self.context.defaults)

    def version(self):
        self.tracker.check_version(self.c)

    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self.version()



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *


class Configure(FiggyTest):
    def __init__(self, **kwargs):
        self.role = kwargs["role"]
        self.run()

    def run(self):
        child = pexpect.spawn(f'figgy --{Utils.get_first(configure)}', timeout=3, encoding='utf-8')
        child.expect('.*Please input.*')
        child.sendline('bastion')
        child.expect('.*What type of user.*')
        child.sendline(self.role)
        child.expect('.*Enable.* colored output.*')
        child.sendline('n')
        child.expect('.*figgy successfully configured..*')



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *
import os


class DataLogin(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(pexpect.spawn(f'{CLI_NAME} login sandbox', timeout=10, encoding='utf-8'), extra_args=extra_args)

    def run(self):
        self.expect('.*user name:.*')
        self.sendline('FiggyDataTester')
        self.expect('.*weirdness.*')
        self.sendline('n')
        self.expect('.*Options.*devops.*')
        self.sendline('data')
        self.expect('.*default account.*')
        self.sendline('stage')
        self.expect('.*Login successful.*')



from dataclasses import dataclass
from typing import Dict


# Todo add support for expiration and check for expiration rather than calling STS with session.
@dataclass
class FiggyAWSSession:
    access_key: str
    secret_key: str
    token: str

    @staticmethod
    def from_sts_response(response: Dict) -> "FiggyAWSSession":
        creds = response.get('Credentials', {})

        return FiggyAWSSession(
            access_key=creds.get('AccessKeyId'),
            secret_key=creds.get('SecretAccessKey'),
            token=creds.get('SessionToken')
        )


"""
Example STS boto response:

{
    'Credentials': {
        'AccessKeyId': 'string',
        'SecretAccessKey': 'string',
        'SessionToken': 'string',
        'Expiration': datetime(2015, 1, 1)
    },
    'AssumedRoleUser': {
        'AssumedRoleId': 'string',
        'Arn': 'string'
    },
    'PackedPolicySize': 123
}
"""



from figcli.config import *


class PluginDao:
    def __init__(self, dynamo_resource):
        self._dynamo_resource = dynamo_resource
        self._plugin_table = self._dynamo_resource.Table(PLUGIN_TABLE_NAME)

    def validate(self, service_name: str, plugin_name: str, sha1: str, user: str):
        """
        Updates a plugin's sha1 in the plugin-state table, which in turn "validates" it.
        Args:
            service_name: service_name as defined in locals.tf
            plugin_name: plugin_name to validate as defined in figgy.json
            sha1: sha1 that is printed as part of the build output
            user: User performing the validation. Needed for change tracking purposes.
        """
        self._plugin_table.update_item(
            Key={
                PLUGIN_SERVICE_KEY_NAME: service_name,
                PLUGIN_PLUGIN_KEY_NAME: plugin_name
            },
            UpdateExpression=f"SET {PLUGIN_SHA_ATTR_NAME}=:sh, {PLUGIN_USER_ATTR_NAME}=:us",
            ExpressionAttributeValues={
                ":sh": sha1,
                ":us": user
            }
        )



import pexpect
from figcli.test.cli.base.configure import Configure
from figcli.utils.utils import *


class Restore:
    def __init__(self, **kwargs):
        self.env = kwargs["env"]
        super().__init__(pexpect.spawn(f'python3 {CLI_NAME} config {Utils.get_first(restore)} --env {self.env}',
                                       encoding='utf-8', timeout=5))

    def choose_key(self, key: str, expect: str) -> None:
        self.expect("Please input.+: ")
        self.sendline(key)
        self.expect(expect)

    def choose_restore_item(self, item_choice: str, expect: str):
        self.sendline(item_choice)
        self.expect(expect)

    def confirm_restore_item(self, expect: str, confirm: str ='y'):
        self.sendline(confirm)
        self.expect(expect)




import time
from enum import Enum

from figcli.config import *
from dataclasses import dataclass
from typing import Dict


class ConfigState(Enum):
    DELETED = 0
    ACTIVE = 1


@dataclass(eq=True, frozen=True)
class ConfigItem:
    name: str
    state: str
    last_updated: int

    @staticmethod
    def from_dict(obj: Dict) -> "ConfigItem":
        name = obj.get(CACHE_PARAMETER_KEY_NAME, None)
        last_updated = obj.get(CACHE_LAST_UPDATED_KEY_NAME, time.time())
        state = obj.get(CACHE_STATE_ATTR_NAME, ConfigState.DELETED.name)

        return ConfigItem(name=name, last_updated=last_updated, state=ConfigState[state])




import time

from figcli.config import *
from datetime import datetime
from typing import List

from botocore.exceptions import ClientError
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from tabulate import tabulate

from figcli.commands.config.delete import Delete
from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.config import ConfigDao
from figcli.data.dao.ssm import SsmDao
from figcli.models.parameter_store_history import PSHistory
from figcli.models.restore_config import RestoreConfig
from figcli.svcs.kms import KmsSvc
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import Utils


class Restore(ConfigCommand):
    def __init__(
            self,
            ssm_init: SsmDao,
            kms_init: KmsSvc,
            config_init: ConfigDao,
            colors_enabled: bool,
            context: ConfigContext,
            config_completer: WordCompleter,
            delete: Delete
    ):
        super().__init__(restore, colors_enabled, context)
        self._config_context = context
        self._ssm = ssm_init
        self._kms = kms_init
        self._config = config_init
        self._utils = Utils(colors_enabled)
        self._point_in_time = context.point_in_time
        self._config_completer = config_completer
        self._delete = delete

    def _client_exception_msg(self, item: RestoreConfig, e: ClientError):
        if "AccessDeniedException" == e.response["Error"]["Code"]:
            print(
                f"\n\nYou do not have permissions to a new config value at the path: "
                f"{self.c.fg_bl}{item.ps_name}{self.c.rs}"
            )
        else:
            print(
                f"{self.c.fg_rd}Error message: "
                f"{e.response['Error']['Message']}{self.c.rs}"
            )

    def get_parameter_arn(self, parameter_name: str):
        account_id = self._ssm.get_parameter(ACCOUNT_ID_PATH)

        return f"arn:aws:ssm:us-east-1:{account_id}:parameter{parameter_name}"

    def _restore_param(self) -> None:
        """
        Allow the user to query a parameter store entry from dynamo, so we can query + restore it, if desired.
        """

        table_entries = []

        ps_name = prompt(f"Please input PS key to restore: ", completer=self._config_completer)

        print(f"Attempting to retrieve all restorable values of {ps_name}")
        items: List[RestoreConfig] = self._config.get_parameter_restore_details(ps_name)

        if len(items) == 0:
            print("No restorable values were found for this parameter.")
            return

        for i, item in enumerate(items):
            date = time.strftime(
                "%Y-%m-%d %H:%M:%S", time.localtime(item.ps_time / 1000)
            )

            # we need to decrypt the value, if encrypted, in order to show it to the user
            if item.ps_key_id:
                item.ps_value = self._kms.decrypt_with_context(
                    item.ps_value,
                    {"PARAMETER_ARN": self.get_parameter_arn(item.ps_name)},
                )
            table_entries.append([i, date, item.ps_value, item.ps_user])

        print(
            tabulate(
                table_entries,
                headers=["Item #", "Time Created", "Value", "User"],
                tablefmt="grid",
                numalign="center",
                stralign="left",
            )
        )

        choice = int(input("Which item number would you like to restore: "))

        item = items[choice] if items[choice] else None

        if not item:
            print(f"{item} wasn't a valid choice from the list. Aborting restore")
            return

        final_selection = input(
            f"Are you sure you want to restore item #{choice} and have it be the latest version? (y/N): "
        )
        if final_selection.lower() != "y":
            print("Okay, aborting restore")
            return

        key_id = None if item.ps_type == "String" else item.ps_key_id

        try:
            self._ssm.set_parameter(
                item.ps_name,
                item.ps_value,
                item.ps_description,
                item.ps_type,
                key_id=key_id,
            )

            current_value = self._ssm.get_parameter(item.ps_name)
            if current_value == item.ps_value:
                print("Restore was successful")
            else:
                print(
                    "Latest version in parameter store doesn't match what we restored."
                )
                print(
                    f"Current value: {current_value}.  Expected value: {item.ps_value}"
                )

        except ClientError as e:
            self._client_exception_msg(item, e)

    def _decrypt_if_applicable(self, entry: RestoreConfig) -> str:
        if entry.ps_type != "String":
            return self._kms.decrypt_with_context(
                entry.ps_value, {"PARAMETER_ARN": self.get_parameter_arn(entry.ps_name)}
            )
        else:
            return entry.ps_value

    def _restore_params_to_point_in_time(self):
        """
        Restores parameters as they were to a point-in-time as defined by the time provided by the users.
        Replays parameter history to that point-in-time so versioning remains intact.
        """

        ps_prefix = prompt(f"Which parameter store prefix would you like to recursively restore? "
                           f"(e.g., /app/demo-time): ", completer=self._config_completer)

        choice = input(f"Are you sure you want to recursively restore from prefix: {ps_prefix}? (y/N): ")
        if choice.lower() != "y":
            print("Aborting restore.")

        time_selected: str = ""

        try:
            time_selected = input("Seconds since epoch to restore latest values from: ")
            time_converted = datetime.fromtimestamp(float(time_selected))
        except ValueError as e:
            if "out of range" in e.args[0]:
                try:
                    time_converted = datetime.fromtimestamp(float(time_selected) / 1000)
                except ValueError as e:
                    print(
                        "Make sure you're using a format of either seconds or milliseconds since epoch."
                    )
                    return
            elif "could not convert" in e.args[0]:
                print(
                    f"The format of this input should be seconds since epoch. (e.g., 1547647091)\n"
                    f"Try using: https://www.epochconverter.com/ to convert your date to this specific format."
                )
                return
            else:
                print(
                    "An unexpected exception triggered: "
                    f"'{e}' while trying to convert {time_selected} to 'datetime' format."
                )
                return

        choice = input(
            f"Are you sure you want to restore latest values from date: {time_converted}? (y/N): "
        )
        if choice.lower() != "y":
            print("Aborting restore.")
            return

        ps_history: PSHistory = self._config.get_parameter_history_before_time(time_converted, ps_prefix)

        if len(ps_history.history.values()) == 0:
            print("No results found for time range.  Aborting.")
            return

        try:
            for item in ps_history.history.values():
                if item.cfg_at(time_converted).ps_action == SSM_PUT:

                    cfgs_before: List[RestoreConfig] = item.cfgs_before(time_converted)
                    cfg_at: RestoreConfig = item.cfg_at(time_converted)
                    ssm_value = self._ssm.get_parameter(item.name)
                    dynamo_value = self._decrypt_if_applicable(cfg_at)

                    if ssm_value != dynamo_value:
                        if ssm_value is not None:
                            self._ssm.delete_parameter(item.name)

                        for cfg in cfgs_before:
                            decrypted_value = self._decrypt_if_applicable(cfg)
                            print(f"Restoring: {cfg.ps_name} as value: {decrypted_value} with description: "
                                  f"{cfg.ps_description} and key: {cfg.ps_key_id if cfg.ps_key_id else 'Parameter was unencrypted'}")
                            print(f"Replaying version: {cfg.ps_version} of {cfg.ps_name}")

                            self._ssm.set_parameter(cfg.ps_name, decrypted_value, cfg.ps_description, cfg.ps_type,
                                                    cfg.ps_key_id)
                    else:
                        print(f"Value for cfg: {item.name} is current. Skipping.")
                else:
                    # This item must have been a delete, which means this config didn't exist at that time.
                    print(f"Checking if {item.name} exists. It was previously deleted.")
                    self._prompt_delete(item.name)
        except ClientError as e:
            self._utils.error_exit(f"Caught error when attempting restore. {e}")

    def _prompt_delete(self, name):
        param = self._ssm.get_parameter_encrypted(name)
        if param:
            selection = "unselected"
            while selection.lower() != "y" and selection.lower() != "n":
                selection = input(f"PS Name: {self.c.fg_bl}{name}{self.c.rs} did not exist at this restore time."
                                  f" Delete it? (y/N)")
                selection = selection if selection != '' else 'n'

            if selection.lower() == 'y':
                self._delete.delete_param(name)

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        if self._point_in_time:
            self._restore_params_to_point_in_time()
        else:
            self._restore_param()



from figcli.models.run_env import RunEnv
from figcli.config import *
import getpass
from decimal import Decimal
from typing import List, Dict, TypeVar
import datetime

T = TypeVar('T', bound='RestoreConfig')


class RestoreConfig:
    """
    Model used for defining values used to restore a parameter from parameter store
    """

    def __init__(
            self,
            ps_description: str,
            ps_name: str,
            ps_time: Decimal,
            ps_type: str,
            ps_key_id: str,
            ps_value: str,
            ps_version: str,
            ps_user: str,
            ps_action: str,
    ):
        self.ps_type = ps_type
        self.ps_key_id = ps_key_id
        self.ps_time = ps_time
        self.ps_description = ps_description
        self.ps_value = ps_value
        self.ps_name = ps_name
        self.ps_version = ps_version
        self.ps_user = ps_user
        self.ps_action = ps_action

        self.props = {
            AUDIT_PARAMETER_ATTR_DESCRIPTION: self.ps_description,
            AUDIT_PARAMETER_KEY_NAME: self.ps_name,
            AUDIT_TIME_KEY_NAME: self.ps_time,
            AUDIT_PARAMETER_ATTR_TYPE: self.ps_type,
            AUDIT_PARAMETER_ATTR_KEY_ID: self.ps_key_id,
            AUDIT_PARAMETER_ATTR_VALUE: self.ps_value,
            AUDIT_PARAMETER_ATTR_VERSION: self.ps_version,
            AUDIT_USER_ATTR_NAME: self.ps_user,
            AUDIT_ACTION_ATTR_NAME: self.ps_action
        }

    @staticmethod
    def convert_to_model(items: List[Dict]) -> List[T]:
        models: List[RestoreConfig] = []

        for item in items:
            model = RestoreConfig(
                item.get(AUDIT_PARAMETER_ATTR_DESCRIPTION, ""),
                item[AUDIT_PARAMETER_KEY_NAME],
                item[AUDIT_TIME_KEY_NAME],
                item.get(AUDIT_PARAMETER_ATTR_TYPE, ""),
                item.get(AUDIT_PARAMETER_ATTR_KEY_ID, None),
                item.get(AUDIT_PARAMETER_ATTR_VALUE, ""),
                item.get(AUDIT_PARAMETER_ATTR_VERSION, ""),
                item.get(AUDIT_USER_ATTR_NAME, ""),
                item[AUDIT_ACTION_ATTR_NAME]
            )

            models.append(model)

        return models

    def __str__(self) -> str:
        return f"{self.__dict__}"



from figcli.models.parameter_history import ParameterHistory
from figcli.models.restore_config import RestoreConfig
from typing import List, Set, Dict


class PSHistory:
    """
    Represents the full history of ParameterStore at a point in time.
    """

    def __init__(self, configs: List[ParameterHistory]):
        self.history: Dict[str, ParameterHistory] = {}
        for cfg in configs:
            self.history[cfg.name] = cfg

    def __str__(self):
        return f"{self.__dict__}"



from typing import Dict, Set

from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.commands.config.delete import Delete
from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.config.style.style import FIGGY_STYLE
from figcli.data.dao.config import ConfigDao
from figcli.data.dao.ssm import SsmDao
from figcli.extras.key_utils import KeyUtils
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *


class Cleanup(ConfigCommand):
    """
    Detects orphaned ParameterStore names, replication configurations, and merge keys, then
    prompts the user to delete them. This is typically run after the `sync` command informs
    the user that there are orphaned configurations.
    """

    def __init__(self, ssm: SsmDao, ddb: ConfigDao, context: ConfigContext,
                 config_completer_init: WordCompleter, colors_enabled: bool, delete: Delete, args=None):
        super().__init__(cleanup, colors_enabled, context)
        self._ssm = ssm  # type: SsmDao
        self._config_dao = ddb  # type: ConfigDao
        self._config_completer = config_completer_init  # type: WordCompleter
        self._utils = Utils(colors_enabled)
        self.example = f"{self.c.fg_bl}{CLI_NAME} config {self.command_printable} --env dev " \
            f"--config /path/to/figgy.json{self.c.rs}"
        self._config_path = context.ci_config_path if context.ci_config_path else Utils.find_figgy_json()

        # If user passes in --info flag, we don't need all of this to be initialized.
        if not hasattr(args, Utils.get_first(info)) or args.info is False:
            # Validate & parse figgy.json
            self._config = self._utils.get_ci_config(self._config_path)  # type: Dict
            self._shared_names = set(self._utils.get_config_key_safe(SHARED_KEY, self._config, default=[]))  # type: Set
            self._repl_conf = self._utils.get_config_key_safe(REPLICATION_KEY, self._config, default={})  # type: Dict
            self._merge_conf = self._utils.get_config_key_safe(MERGE_KEY, self._config, default={})  # type: Dict
            self._config_keys = set(self._utils.get_config_key_safe(CONFIG_KEY, self._config, default=[]))  # type: Set
            self._merge_keys = set(self._merge_conf.keys())  # type: Set
            self._namespace = self._utils.get_namespace(self._config)  # type: str
            self._delete_command = delete
            self._repl_from_conf = self._utils.get_config_key_safe(REPL_FROM_KEY, self._config, default={})
            self._repl_conf = KeyUtils.merge_repl_and_repl_from_blocks(self._repl_conf, self._repl_from_conf,
                                                                       self._namespace)

            # Build list of all keys found across all config types
            self._all_keys = KeyUtils().find_all_expected_names(self._config_keys, self._shared_names, self._merge_conf,
                                                                self._repl_conf, self._repl_from_conf, self._namespace)

    # Prompts for this file
    def _cleanup_parameters(self, config_keys: Set):
        """
        Prompts user for cleanup of orphaned ParameterStore names.
        Args:
            config_keys: set() -> Set of parameters that are found as defined in the figgy.json file for a svc
        """

        print(f"{self.c.fg_gr}Checking for orphaned config names.{self.c.rs}\r\n")
        # Find & Cleanup orphaned keys
        ps_keys = set(list(map(lambda x: x['Name'], self._ssm.get_all_parameters([self._namespace]))))
        ps_only_keys = ps_keys.difference(config_keys)
        for key in ps_only_keys:
            selection = "unselected"
            while selection.lower() != "y" and selection.lower() != "n":
                exists_msg = [
                    (f'class:{self.c.bl}', key),
                    ('class:', ' exists in ParameterStore but does not exist '
                               'in your config, do you want to delete it? (y/N): ')
                ]
                selection = prompt(exists_msg, completer=WordCompleter(['Y', 'N']), style=FIGGY_STYLE)
                selection = selection if selection != '' else 'n'
                if selection.strip().lower() == "y":
                    self._delete_command.delete_param(key)
        if not ps_only_keys:
            print(f"{self.c.fg_bl}No orphaned keys found.{self.c.rs}")

    def _cleanup_replication(self) -> None:
        """
        Cleans up orphaned replication and merge configurations.
        Args:
            config_repl: The replication config dictionary as parsed from the figgy.json file
            shared_names: Expected parameters as defined in the figgy.json
            config_merge: The merge config dict as defined
            run_env: RunEnv object
            namespace: str -> /app/service-name as defined or parsed from the figgy.json file.
        """

        print(f"{self.c.fg_gr}Checking for orphaned replication configs.{self.c.rs}")
        remote_cfgs = self._config_dao.get_all_configs(self.run_env, self._namespace)
        notify = True
        if remote_cfgs:
            for cfg in remote_cfgs:
                if cfg.source not in list(self._repl_conf.keys()) \
                        and cfg.destination not in list(self._repl_conf.values()) \
                        and cfg.destination not in self._shared_names \
                        and cfg.destination not in list(self._merge_conf.keys()) \
                        and (isinstance(cfg.source, list) or cfg.source.startswith(shared_ns)
                             or cfg.source.startswith(self.context.defaults.service_ns)):
                    notify = False
                    selection = "unselected"
                    while selection.lower() != "y" and selection.lower() != "n":
                        selection = input(
                            f"Remote replication config with {self.c.fg_bl}{self._namespace}{self.c.rs} replication "
                            f"mapping of: {self.c.fg_bl}{cfg.source} -> {cfg.destination}{self.c.rs} does not "
                            f"exist in your figgy.json. Should this be removed? (y/N): ").lower()
                        selection = selection if selection != '' else 'n'
                        if selection == "y":
                            self._config_dao.delete_config(cfg.destination, self.run_env)
        if notify:
            print(
                f"{self.c.fg_bl}No remote replication configs found available for cleanup under namespace: {self.c.rs}"
                f"{self.c.fg_gr}{self._namespace}{self.c.rs}")

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        # cleanup service configs
        print()
        self._cleanup_parameters(set(self._all_keys))

        print()
        # cleanup replication configs
        self._cleanup_replication()



import jsonpickle
import uuid
import os
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.provider import Provider
from figcli.models.defaults.provider_config import ProviderConfig, ProviderConfigFactory, BastionProviderConfig
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.config.constants import APP_NS_OVERRIDE
from figcli.config import *
from typing import Dict, Optional, List, Any
from dataclasses import dataclass, field
from figcli.utils.utils import Utils


@dataclass
class CLIDefaults:
    """
    Defaults are parsed from the ~/.figgy/devops/defaults file and then hydrate this object
    """
    user_id: Optional[str]
    role: Optional[Role]
    colors_enabled: bool
    run_env: RunEnv
    region: str
    mfa_enabled: bool
    service_ns: str
    provider: Provider
    session_duration: int
    usage_tracking: bool
    report_errors: Optional[bool]
    auto_mfa: Optional[bool]
    provider_config: Optional[Any]
    mfa_serial: Optional[str]
    user: Optional[str]
    valid_envs: Optional[List[RunEnv]] = field(default_factory=list)
    valid_roles: Optional[List[Role]] = field(default_factory=list)
    assumable_roles: Optional[List[AssumableRole]] = field(default_factory=list)
    extras: Optional[Dict] = field(default_factory=dict)

    def __str__(self):
        return f"Role: {self.role}\nColors Enabled: {self.colors_enabled}\nOkta User: {self.user}\n" \
               f"Default Environment: {self.run_env}"

    @staticmethod
    def unconfigured():
        return CLIDefaults(role=Role("unconfigured"),
                           colors_enabled=False,
                           user=None,
                           run_env=RunEnv("unconfigured"),
                           provider=Provider.UNSELECTED,
                           session_duration=DEFAULT_SESSION_DURATION,
                           region="us-east-1",
                           mfa_enabled=False,
                           mfa_serial=None,
                           provider_config=None,
                           report_errors=False,
                           auto_mfa=False,
                           user_id=str(uuid.uuid4()),
                           service_ns="/app",
                           usage_tracking=False,
                           extras={})

    @staticmethod
    def sandbox(user: str, role: str, colors: bool):
        return CLIDefaults(role=Role(role),
                           colors_enabled=colors,
                           user=user,
                           run_env=RunEnv("unconfigured"),
                           provider=Provider.AWS_BASTION,
                           session_duration=SANDBOX_SESSION_DURATION,
                           region=FIGGY_SANDBOX_REGION,
                           mfa_enabled=False,
                           mfa_serial=None,
                           provider_config=BastionProviderConfig(FIGGY_SANDBOX_PROFILE),
                           report_errors=False,
                           auto_mfa=False,
                           user_id=str(uuid.uuid4()),
                           service_ns="/app",
                           usage_tracking=True,
                           extras={DISABLE_KEYRING: True})

    @staticmethod
    def from_profile(profile):
        return CLIDefaults(role=Role(profile),
                           colors_enabled=Utils.not_windows(),
                           user=profile,
                           run_env=RunEnv(profile),
                           provider=Provider.PROFILE,
                           session_duration=DEFAULT_SESSION_DURATION,
                           region="us-east-1",
                           mfa_enabled=False,
                           mfa_serial=None,
                           provider_config=None,
                           report_errors=False,
                           auto_mfa=False,
                           user_id=str(uuid.uuid4()),
                           service_ns=os.environ.get(APP_NS_OVERRIDE) or "/app",
                           usage_tracking=True,
                           extras={})

    def __str__(self) -> str:
        return jsonpickle.encode(self)



import argparse
import getpass
import logging
import os
import sys
import boto3

from json import JSONDecodeError
from typing import Optional, List

from figcli.svcs.setup import FiggySetup

from figcli.config import *
from figcli.input.input import Input
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.observability.error_reporter import FiggyErrorReporter
from figcli.svcs.auth.provider.provider_factory import SessionProviderFactory
from figcli.svcs.auth.session_manager import SessionManager
from figcli.utils.utils import Utils

from figcli.commands.command_factory import CommandFactory
from figcli.commands.figgy_context import FiggyContext
from figcli.commands.types.command import Command

root_logger = logging.getLogger()
root_logger.setLevel(logging.CRITICAL)
root_logger.handlers = []
stdout_handler = logging.StreamHandler(sys.stdout)

log = logging.getLogger(__name__)


class FiggyCLI:
    @staticmethod
    def add_arg(com_parser, com_arg, cmd, rsc):
        com_parser.add_argument(f'--{Utils.get_first(com_arg)}', help=HELP_TEXT_MAP[com_arg],
                                action=arg_options[rsc][cmd][com_arg][action],
                                required=arg_options[rsc][cmd][com_arg][required])

    @staticmethod
    def parse_args():
        """
        Parses Figgy command line arguments and returns generic "args" object.
        """
        parser = argparse.ArgumentParser(description=RESOURCE_PARSER_DESC)
        parser.add_argument(f'--{Utils.get_first(configure)}', help=CONFIGURE_HELP_TEXT, action=store_true)
        parser.add_argument(f'--{Utils.get_first(prompt_com)}', help=PROMPT_HELP_TEXT, action=store_true)
        parser.add_argument(f'--{Utils.get_first(version)}', help=VERSION_HELP_TEXT, action=store_true)
        parser.add_argument(f'--{Utils.get_first(skip_upgrade)}', help=SKIP_UPGRADE_HELP_TEXT, action=store_true)
        parser.add_argument(f'--{Utils.get_first(upgrade)}', help=UPGRADE_HELP_TEXT, action=store_true)

        resource_subparsers = parser.add_subparsers(title='resources', dest='resource', metavar='')

        for rsc in resource_map:
            cmd_parser = resource_subparsers.add_parser(Utils.get_first(rsc), help=HELP_TEXT_MAP[rsc])
            subparser = cmd_parser.add_subparsers(title=f'{Utils.get_first(rsc)} commands', dest='command', metavar='',
                                                  help=HELP_TEXT_MAP[rsc])

            for cmd in resource_map[rsc]:
                com_parser = subparser.add_parser(Utils.get_first(cmd), help=HELP_TEXT_MAP[cmd])
                for com_arg, val in arg_options[rsc][cmd].items():
                    FiggyCLI.add_arg(com_parser, com_arg, cmd, rsc)

        return parser.parse_args()

    def get_profile(self, prompt: bool) -> str:
        """Returns the user's profile.

        Checks ENV variable, if not there, checks the config file (created via the --configure option), otherwise prompts
        the user

        Args:
            prompt: True/False - if True, users will always be prompted to input their profile

        :return: str: aws profile name
        """

        if BASTION_PROFILE_ENV_NAME in os.environ and not prompt:
            return os.environ.get(BASTION_PROFILE_ENV_NAME)
        else:
            defaults: CLIDefaults = FiggySetup.stc_get_defaults(self._is_setup_command, profile=self._profile)
            if defaults is not None and not prompt:
                return defaults.provider_config.profile
            else:
                return Input.select_aws_cli_profile()

    def get_role(self, prompt: bool, role_override: str = None) -> Role:
        """
        Returns a string of the user's selected role.

        Lookup the user's default role from the config file (created via the --configure option), an ENV variable, or
        instead prompt the user for the session.

        :param prompt: True/False - if True, users will always be prompted to input their role
        :param role_override: String representation of the role to get, regardless of defaults.
        :return: str: name of the selected role.
        """

        defaults = FiggySetup.stc_get_defaults(self._is_setup_command, profile=self._profile)
        if defaults is not None and not prompt:

            if role_override:
                if role_override in [role.role.role for role in defaults.assumable_roles]:
                    print(f'Role override selected: {role_override}\n')
                    return Role(role_override)
                else:
                    self._utils.error_exit(f"Invalid role override provided of: {role_override}. "
                                           f"You do not have permissions to assume this role. Contact your system "
                                           f"administrator to receive permissions then rerun `{CLI_NAME} "
                                           f"--{Utils.get_first(configure)}`.")

            return defaults.role
        else:
            roles = self.__setup().get_assumable_roles()
            role_names = list(set([x.role.role for x in roles]))
            return Input.select_role(role_names)

    def get_colors_enabled(self) -> bool:
        """
        Defaults to true, unless user ran --configure and disabled colored output
        Returns: True/False
        """

        defaults = FiggySetup.stc_get_defaults(skip=self._is_setup_command, profile=self._profile)
        if defaults is not None:
            return defaults.colors_enabled
        else:
            return Utils.not_windows()

    def get_command(self) -> Command:
        """
        Maps the user's passed in text command to one of our defined 'command' objects we use in the code.
        Args:

        Returns: command object.
        """
        return self.get_command_factory().instance()

    def find_assumable_role(self, env: RunEnv, role: Role, skip: bool = False, profile=None) -> AssumableRole:
        """
        Looks up the appropriate assumable role based on the user's selected defaults or command-line overrides for
        --env, --role, and --profile.
        """

        if profile:
            return AssumableRole.from_profile(profile)

        assumable_roles: List[AssumableRole] = FiggySetup.stc_get_defaults(skip=skip).assumable_roles
        matching_role = [ar for ar in assumable_roles if ar.role == role and ar.run_env == env]
        if matching_role:
            matching_role = matching_role.pop()
        else:
            matching_role = None

        return matching_role

    def __setup(self) -> FiggySetup:
        if not self._setup:
            self._setup = FiggySetup()

        return self._setup

    @staticmethod
    def is_setup_command(args):
        return Utils.is_set_true(configure, args) \
               or Utils.command_set(sandbox, args) \
               or Utils.is_set_true(version, args) \
               or Utils.attr_exists(profile, args) \
               or Utils.is_set_true(upgrade, args)

    def __init__(self, args):
        """
        Initializes global shared properties
        :param args: Arguments passed in from user, collected from ArgParse
        """
        self._mgmt_session = None
        self._s3_resource = None
        self._mgmt_ssm = None
        self._profile = None
        self._command_factory = None
        self._setup = None
        self._session_manager = None
        self._is_setup_command: bool = FiggyCLI.is_setup_command(args)
        self._utils = Utils(self.get_colors_enabled())
        self._sts = boto3.client('sts')
        self._profile = Utils.attr_if_exists(profile, args)
        self._defaults: CLIDefaults = FiggySetup.stc_get_defaults(skip=self._is_setup_command, profile=self._profile)
        self._run_env = self._defaults.run_env
        role_override = Utils.attr_if_exists(role, args)
        self._role: Role = self.get_role(args.prompt, role_override=role_override)

        if not self._is_setup_command:
            if not hasattr(args, 'env') or args.env is None:
                print(f"{EMPTY_ENV_HELP_TEXT}{self._run_env.env}\n")
            else:
                Utils.stc_validate(args.env in self._defaults.valid_envs,
                                   f'{ENV_HELP_TEXT} {self._defaults.valid_envs}. Provided: {args.env}')
                self._run_env = RunEnv(args.env)

        self._utils.validate(Utils.attr_exists(configure, args) or Utils.attr_exists(command, args),
                             f"No command found. Proper format is `{CLI_NAME} <resource> <command> --option(s)`")

        self._assumable_role = self.find_assumable_role(self._run_env, self._role, skip=self._is_setup_command,
                                                        profile=self._profile)
        # Todo validate this role?

        command_val = Utils.attr_if_exists(command, args)
        resource_val = Utils.attr_if_exists(resource, args)
        found_command: frozenset = frozenset({Utils.attr_if_exists(command, args)}) if command_val else None
        found_resource: frozenset = frozenset({Utils.attr_if_exists(resource, args)}) if resource_val else None

        self._context = FiggyContext(self.get_colors_enabled(), found_resource, found_command,
                                     self._run_env, self._assumable_role, args)

    def get_command_factory(self) -> CommandFactory:
        if not self._command_factory:
            self._command_factory = CommandFactory(self._context,
                                                   FiggySetup.stc_get_defaults(skip=self._is_setup_command,
                                                                               profile=self._profile))

        return self._command_factory


def main():
    """
        Entrypoint to figgy.

        Performs generic validation, then routes user down appropriate execution path based on command line parameters
    """
    arguments = sys.argv
    user = getpass.getuser()
    Utils.stc_validate(user != ROOT_USER, f"Hey! Stop trying to run {CLI_NAME} as {ROOT_USER}. That's bad!")
    original_command = ' '.join(arguments)
    sys.argv = arguments
    try:
        # Parse / Validate Args
        args = FiggyCLI.parse_args()
        if hasattr(args, 'debug') and args.debug:
            root_logger.setLevel(logging.INFO)
            root_logger.addHandler(stdout_handler)

        cli: FiggyCLI = FiggyCLI(args)
        command: Command = cli.get_command()

        if hasattr(args, 'info') and args.info:
            command.print_help_text()
        else:
            command.execute()

    except AssertionError as e:
        Utils.stc_error_exit(e.args[0])
    except Exception as e:
        try:
            error_reporter = FiggyErrorReporter(FiggySetup.stc_get_defaults(skip=True, profile=None))
            error_reporter.log_error(original_command, e)
        except Exception as e:
            print(e)
            print(f"\n\nUnable to log or report this exception. Please submit a Github issue to: {FIGGY_GITHUB}")
    except KeyboardInterrupt:
        exit(1)


if __name__ == '__main__':
    try:
        main()
    except Warning:
        pass



from botocore.exceptions import ClientError
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.svcs.auth.session_manager import SessionManager
from figcli.utils.utils import *


# Todo: FIX - Prompt user for next environment since we can no longer assume what the next environment is
class Promote(ConfigCommand):

    def __init__(self, source_ssm: SsmDao, config_completer_init: WordCompleter,
                 colors_enabled: bool, config_context: ConfigContext, session_mgr: SessionManager):
        super().__init__(promote, colors_enabled, config_context)
        self.config_context = config_context
        self._source_ssm = source_ssm
        self._session_mgr = session_mgr
        self._config_completer = config_completer_init
        self._utils = Utils(colors_enabled)


    def _promote(self):
        repeat = True
        parameters: List[Dict] = []
        while repeat:
            namespace = prompt(f"Please input a namespace prefix to promote:"
                               f" (i.e. {self.context.defaults.service_ns}/foo/): ",
                               completer=self._config_completer)
            if not self._utils.is_valid_input(namespace, "namespace", notify=False):
                continue

            try:
                parameters: List[Dict] = self._source_ssm.get_all_parameters([namespace])

                if not parameters and self._source_ssm.get_parameter(namespace):
                    parameters = [self._source_ssm.get_parameter_details(namespace)]

                if parameters:
                    repeat = False
                else:
                    print("No parameters found. Try again.\n")
            except ClientError as e:
                print(f"{self.c.fg_rd}ERROR: >> {e}{self.c.rs}")
                continue

        print(f'{self.c.fg_bl}Found {len(parameters)} parameters to migrate.{self.c.rs}\n')

        assumable_roles = self.context.defaults.assumable_roles
        matching_roles = list(set([x for x in assumable_roles if x.role == self.config_context.role]))
        valid_envs = set([x.run_env.env for x in matching_roles])
        valid_envs.remove(self.run_env.env)  # Remove current env, we can't promote from dev -> dev
        next_env = Input.select(f'Please select the destination environment.', valid_options=list(valid_envs))

        matching_role = [role for role in matching_roles if role.run_env == RunEnv(next_env)][0]
        dest_ssm = SsmDao(self._session_mgr.get_session(matching_role, prompt=False).client('ssm'))

        for param in parameters:
            if 'KeyId' in param:
                print(f"Skipping param: {self.c.fg_bl}{param['Name']}{self.c.rs}. "
                      f"It is encrypted and cannot be migrated.")
            else:
                promote_it = Input.y_n_input(f"Would you like to promote: {param['Name']}?",
                                             default_yes=True)

                if promote_it:
                    val = self._source_ssm.get_parameter(param['Name'])
                    description = param.get('Description', "")
                    dest_ssm.set_parameter(param['Name'], val, description, SSM_STRING)
                    print(f"Successfully promoted {self.c.fg_bl}{param['Name']}{self.c.rs} to {next_env}.\r\n")


    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._promote()



import logging
import os

from typing import Callable, List, Dict, Optional

from tabulate import tabulate
from figcli.config import *
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.models.assumable_role import AssumableRole
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.defaults.provider import Provider
from figcli.models.defaults.provider_config import ProviderConfigFactory
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.config_manager import ConfigManager
from figcli.svcs.auth.provider.provider_factory import SessionProviderFactory
from figcli.svcs.auth.provider.session_provider import SessionProvider
from figcli.svcs.auth.provider.sso_session_provider import SSOSessionProvider
from figcli.svcs.auth.session_manager import SessionManager
from figcli.utils.secrets_manager import SecretsManager
from figcli.utils.utils import Utils
from json import JSONDecodeError

log = logging.getLogger(__name__)


class FiggySetup:
    """
    Contains logic around setting up Figgy. Configuring user auth, etc.
    """

    # If we ever need to add params to this constructor we'll need to better handle dependencies and do a bit of
    # refactoring here.
    def __init__(self):
        self._cache_mgr = CacheManager(file_override=DEFAULTS_FILE_CACHE_PATH)
        self._config_mgr, self.c = ConfigManager.figgy(), Utils.default_colors()
        self._session_mgr = None
        self._session_provider = None

    def get_assumable_roles(self, defaults: CLIDefaults = None) -> List[AssumableRole]:
        if not defaults:
            defaults = self.get_defaults()

        return self._get_session_provider(defaults).get_assumable_roles()

    def _get_session_manager(self, defaults: CLIDefaults) -> SessionManager:
        if not self._session_mgr:
            self._session_mgr = SessionManager(defaults, self._get_session_provider(defaults))

        return self._session_mgr

    def _get_session_provider(self, defaults: CLIDefaults):
        if not self._session_provider:
            self._session_provider = SessionProviderFactory(defaults).instance()

        return self._session_provider

    def configure_auth(self, current_defaults: CLIDefaults, configure_provider=True) -> CLIDefaults:
        updated_defaults = current_defaults
        if configure_provider or current_defaults.provider is Provider.UNSELECTED:
            provider: Provider = Input.select_provider()
            updated_defaults.provider = provider
        else:
            provider: Provider = current_defaults.provider

        if provider in Provider.sso_providers():
            user: str = Input.get_user(provider=provider.name)
            password: str = Input.get_password(provider=provider.name)
            SecretsManager.set_password(user, password)
            updated_defaults.user = user

        try:
            mfa_enabled = Utils.parse_bool(self._config_mgr.get_or_prompt(Config.Section.Figgy.MFA_ENABLED,
                                                                          Input.select_mfa_enabled))
            if mfa_enabled:
                auto_mfa = Utils.parse_bool(self._config_mgr.get_or_prompt(Config.Section.Figgy.AUTO_MFA,
                                                                           Input.select_auto_mfa))
            else:
                auto_mfa = False

        except ValueError as e:
            Utils.stc_error_exit(f"Invalid value found in figgy defaults file under "
                                 f"{Config.Section.Figgy.MFA_ENABLED.value}. It must be either 'true' or 'false'")
        else:
            updated_defaults.mfa_enabled = mfa_enabled
            updated_defaults.auto_mfa = auto_mfa

        if updated_defaults.auto_mfa:
            mfa_secret = Input.get_mfa_secret()
            SecretsManager.set_mfa_secret(updated_defaults.user, mfa_secret)

        if configure_provider:
            provider_config = ProviderConfigFactory().instance(provider, mfa_enabled=updated_defaults.mfa_enabled)
            updated_defaults.provider_config = provider_config

        return updated_defaults

    def configure_roles(self, current_defaults: CLIDefaults, run_env: RunEnv = None, role: Role = None) -> CLIDefaults:
        updated_defaults = current_defaults
        provider_factory: SessionProviderFactory = SessionProviderFactory(current_defaults)
        session_provider: SSOSessionProvider = provider_factory.instance()
        session_provider.cleanup_session_cache()

        # Get assertion and parse out account -> role -> run_env mappings.
        assumable_roles: List[AssumableRole] = session_provider.get_assumable_roles()
        print(
            f"\n{self.c.fg_bl}The following roles were detected for user: {current_defaults.user} - if something is missing, "
            f"contact your system administrator.{self.c.rs}\n")

        if assumable_roles:
            self.print_role_table(assumable_roles)

        valid_envs = list(set([x.run_env.env for x in assumable_roles]))
        valid_roles = list(set([x.role.role for x in assumable_roles]))

        if not role:
            role: Role = Input.select_role(valid_roles=valid_roles)
            print("\n")

        if not run_env:
            run_env: RunEnv = Input.select_default_account(valid_envs=valid_envs)
            print("\n")

        updated_defaults.run_env = run_env
        updated_defaults.valid_envs = valid_envs
        updated_defaults.valid_roles = valid_roles
        updated_defaults.assumable_roles = assumable_roles
        updated_defaults.role = role

        return updated_defaults

    def configure_preferences(self, current_defaults: CLIDefaults):
        updated_defaults = current_defaults
        updated_defaults.region = self._config_mgr.get_or_prompt(Config.Section.Figgy.AWS_REGION, Input.select_region)
        updated_defaults.colors_enabled = self._config_mgr.get_or_prompt(Config.Section.Figgy.COLORS_ENABLED,
                                                                         Input.select_enable_colors)

        # Defaulting to True, users will always be prompted to report or not report an error.
        updated_defaults.report_errors = True

        updated_defaults.usage_tracking = self._config_mgr.get_or_prompt(Config.Section.Figgy.USAGE_TRACKING,
                                                                         Input.select_usage_tracking)
        return updated_defaults

    def configure_extras(self, current_defaults: CLIDefaults):
        updated_defaults = current_defaults
        if os.environ.get(FIGGY_DISABLE_KEYRING) == 'true':
            updated_defaults.extras[DISABLE_KEYRING] = True

        return updated_defaults

    def configure_figgy_defaults(self, current_defaults: CLIDefaults):
        updated_defaults = current_defaults
        session = self._get_session_manager(current_defaults).get_session(current_defaults.assumable_roles[0],
                                                                          prompt=True)
        ssm = SsmDao(session.client('ssm'))
        default_service_ns = ssm.get_parameter(PS_FIGGY_DEFAULT_SERVICE_NS_PATH)
        updated_defaults.service_ns = default_service_ns

        return updated_defaults

    def basic_configure(self, configure_provider=True) -> CLIDefaults:
        defaults: CLIDefaults = self.get_defaults()
        if not defaults:
            Utils.stc_error_exit(f"Please run {CLI_NAME} --{Utils.get_first(configure)} to set up Figgy, "
                                 f"you've got problems friend!")
        else:
            defaults = self.configure_auth(defaults, configure_provider=configure_provider)

        return defaults

    def save_defaults(self, defaults: CLIDefaults):
        self._cache_mgr.write(DEFAULTS_KEY, defaults)

    def get_defaults(self) -> CLIDefaults:
        try:
            last_write, defaults = self._cache_mgr.get(DEFAULTS_KEY)
        except Exception as e:
            # If cache is corrupted or inaccessible, "fogetaboutit" (in italian accent)
            return CLIDefaults.unconfigured()

        return defaults if defaults else CLIDefaults.unconfigured()

    @staticmethod
    def stc_get_defaults(skip: bool = False, profile: str = None) -> Optional[CLIDefaults]:
        """Lookup a user's defaults as configured by --configure option.
        :param skip - Boolean, if this is true, exit and return none.
        :param profile - AWS CLI profile to use as an override. If this is passed in all other options are ignored.
        :return: hydrated CLIDefaults object of default values stored in cache file or None if no cache found
        """
        if profile:
            return CLIDefaults.from_profile(profile)

        cache_mgr = CacheManager(file_override=DEFAULTS_FILE_CACHE_PATH)
        try:
            last_write, defaults = cache_mgr.get(DEFAULTS_KEY)

            if not defaults:
                if skip:
                    return CLIDefaults.unconfigured()
                else:
                    Utils.stc_error_exit(f'{CLI_NAME} has not been configured. '
                                     f'Please run {CLI_NAME} --{Utils.get_first(configure)}')

            return defaults
        except JSONDecodeError:
            return None

    @staticmethod
    def print_role_table(roles: List[AssumableRole]):
        printable_roles: Dict[int: Dict] = {}
        for role in roles:
            item = printable_roles.get(role.account_id, {})
            item['env'] = role.run_env.env
            item['roles'] = item.get('roles', []) + [role.role.role]
            printable_roles[role.account_id] = item

        print(tabulate(
            [
                [
                    f'{account_id[0:5]} [REDACTED]',
                    printable_roles[account_id]['env'],
                    ', '.join(printable_roles[account_id]['roles'])
                ]
                for account_id in printable_roles.keys()
            ],
            headers=['Account #', 'Environment', 'Roles'],
            tablefmt="grid",
            numalign="center",
            stralign="left",
        ))



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *


class DataPut(FiggyTest):
    def __init__(self, extra_args=""):
        super().__init__(pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(put)} --env {DEFAULT_ENV} '
                                       f'--skip-upgrade {extra_args}',
                      timeout=7, encoding='utf-8'), extra_args=extra_args)
        self.step(f"Testing `{CLI_NAME} config {Utils.get_first(put)} --env {DEFAULT_ENV}`")

    def run(self):
        self.expect('.*Please input a PS Name.*')
        self.sendline(data_param_1)
        self.expect('.*Please input a value.*')
        self.sendline(data_param_1_val)
        self.expect('.*Please input an optional.*')
        self.sendline(data_param_1_desc)
        self.expect('.*secret?.*')
        self.sendline('n')
        self.expect('.*Add another?.*')
        self.sendline('y')
        self.expect('.*PS Name.*')
        self.sendline('/devops/test/invalid')
        self.expect('.*value.*')
        self.sendline(DELETE_ME_VALUE)
        self.expect('.*not have permissions.*another?.*')
        self.sendline('n')

    def add(self, key, value, desc, add_more=False):
        self.expect('.*Please input a PS Name.*')
        self.sendline(key)
        self.expect('.*Please input a value.*')
        self.sendline(value)
        self.expect('.*Please input an optional.*')
        self.sendline(desc)
        self.expect('.*secret?.*')
        self.sendline('n')
        self.expect('.*another.*')
        if add_more:
            self.sendline('y')
        else:
            self.sendline('n')

    def add_another(self, key, value, desc, add_more=True):
        print(f"Adding another: {key} -> {value}")
        self.expect('.*PS Name.*')
        self.sendline(key)
        self.expect('.*Please input a value.*')
        self.sendline(value)
        self.expect('.*Please input an optional.*')
        self.sendline(desc)
        self.expect('.*secret?.*')
        self.sendline('n')
        self.expect('.*Add another.*')
        if add_more:
            self.sendline('y')
        else:
            self.sendline('n')



from typing import Union

from figcli.commands.config_context import ConfigContext
from figcli.commands.help_context import HelpContext
from figcli.config import *
from figcli.config.style.terminal_factory import TerminalFactory
from figcli.utils.utils import Utils
from abc import ABC, abstractmethod
from figcli.commands.command_context import CommandContext


class Command(ABC):
    """
    Root command class from which all other command classes inherit.
    """

    def __init__(self, command_type: frozenset, colors_enabled: bool,
                 context: Union[CommandContext, HelpContext, ConfigContext]):
        self.type = command_type
        self.run_env = context.run_env
        self.c = TerminalFactory(colors_enabled).instance().get_colors()
        self.command_printable = list(self.type)[0]
        self.context = context
        self.example = f"{self.c.fg_bl}{CLI_NAME} {Utils.get_first(context.resource)} {self.command_printable} " \
                       f"--env dev{self.c.rs}"

    @abstractmethod
    def execute(self):
        pass

    def print_help_text(self):
        """
        This help text is printed when users pass in the --info option
        :param command: one of the define commands in config.py
        """

        optional_params = frozenset(filter(lambda x: not arg_options[self.context.resource][self.type][x][required],
                                           arg_options[self.context.resource][self.type].keys()))

        required_params = frozenset(filter(lambda x: arg_options[self.context.resource][self.type][x][required],
                                           arg_options[self.context.resource][self.type].keys()))

        print(f"Command: {self.c.fg_bl}{self.command_printable}{self.c.rs}")
        if len(required_params) > 0:
            print(f"Required parameters: {self.c.fg_bl}{CollectionUtils.printable_set(required_params)}{self.c.rs}")

        if len(optional_params) > 0:
            print(f"Optional parameters: {self.c.fg_bl}{CollectionUtils.printable_set(optional_params)}{self.c.rs}")

        print(f"Example: {self.example}")
        print(f"\n\n{HELP_TEXT_MAP[self.type]}\n")



import sys

import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.dev.get import DevGet
from figcli.test.cli.dev.delete import DevDelete
from figcli.test.cli.dev.put import DevPut

from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *
import uuid
import time


class DevSync(FiggyTest):

    def __init__(self, extra_args=""):
        super().__init__(None, extra_args=extra_args)
        self.missing_key = '/app/ci-test/v1/config12'

    def run(self):
        self.step("Preparing sync process...")
        self.prep_sync()
        self.step("Testing successful sync")
        self.sync_success()
        self.step("Testing multi-level sync")
        self.sync_multi_level_success()
        self.step("Testing sync with known orphaned parameters.")
        self.sync_with_orphans()

    def prep_sync(self):
        put = DevPut(extra_args=self.extra_args)
        put.add('/app/ci-test/v1/config9', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/app/ci-test/v1/config11', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl2', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl3', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl4', DELETE_ME_VALUE, desc='desc', add_more=False)
        print("EXECUTING DELETE!!!!")
        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(self.missing_key)

    def sync_success(self):
        print(f"Testing: {CLI_NAME} config {Utils.get_first(sync)} --env {DEFAULT_ENV} "
              f"--config figcli/test/assets/success/figgy.json")
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(sync)} --env {DEFAULT_ENV} '
                              f'--config figcli/test/assets/success/figgy.json --skip-upgrade {self.extra_args}',
                              encoding='utf-8', timeout=10)
        child.logfile = sys.stdout
        missing_key = '/app/ci-test/v1/config12'
        child.expect(f'.*Please input a value for.*{missing_key}.*')
        child.sendline(DELETE_ME_VALUE)
        child.expect('.*optional description:.*')
        child.sendline('desc')
        child.expect('.*value a secret.*')
        child.sendline('n')
        child.expect('.*Sync completed with no errors!')

    def sync_multi_level_success(self):
        missing_key = '/app/dev/thing/ci-test2/app/ci-test/v1/config12'
        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(missing_key)

        put = DevPut(extra_args=self.extra_args)
        put.add('/app/dev/thing/ci-test2/config9', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/app/dev/thing/ci-test2/config11', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl2', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl3', DELETE_ME_VALUE, desc='desc', add_more=True)
        put.add('/shared/jordan/testrepl4', DELETE_ME_VALUE, desc='desc', add_more=False)

        print(f"Testing: {CLI_NAME} config {Utils.get_first(sync)} --env {DEFAULT_ENV} "
              f"--config figcli/test/assets/success/multi-level-ns.json")
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(sync)} --env {DEFAULT_ENV} '
                              f'--config figcli/test/assets/success/multi-level-ns.json --skip-upgrade {self.extra_args}',
                              encoding='utf-8', timeout=10)

        child.expect(f'.*Please input a value for.*{missing_key}.*')
        child.sendline(DELETE_ME_VALUE)
        child.expect('.*optional description:.*')
        child.sendline('desc')
        child.expect('.*value a secret.*')
        child.sendline('n')
        child.expect('.*Sync completed with no errors!')

    def sync_with_orphans(self):
        delete = DevDelete(extra_args=self.extra_args)
        delete.delete(self.missing_key)
        print("Successful sync + cleanup passed!")

        print(f"Testing: {CLI_NAME} config {Utils.get_first(sync)} --env {DEFAULT_ENV} "
              f"--config figcli/test/assets/error/figgy.json")
        child = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(sync)} --env {DEFAULT_ENV} '
                              f'--config figcli/test/assets/error/figgy.json --skip-upgrade {self.extra_args}', timeout=10)
        child.expect('.*Unused Parameter:.*/app/ci-test/v1/config11.*Orphaned replication.*/shared/jordan/testrepl2.*')
        print("Sync with orphaned configs passed!")



import json
import logging
import os
import uuid
from typing import Set, List, Tuple, Optional

import boto3
from botocore.exceptions import NoCredentialsError, ClientError, ParamValidationError

from figcli.config.constants import *
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.models.assumable_role import AssumableRole
from figcli.models.aws_session import FiggyAWSSession
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.defaults.provider import Provider
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.auth.provider.session_provider import SessionProvider
from figcli.svcs.vault import FiggyVault
from figcli.utils.secrets_manager import SecretsManager
from figcli.utils.utils import Utils, InvalidSessionError
import time

log = logging.getLogger(__name__)


class BastionSessionProvider(SessionProvider):
    _MAX_ATTEMPTS = 5

    def __init__(self, defaults: CLIDefaults):
        super().__init__(defaults)
        self.__id = uuid.uuid4()
        self._utils = Utils(defaults.colors_enabled)
        self.__bastion_session = boto3.session.Session(profile_name=self._defaults.provider_config.profile_name)
        self._ssm = None
        self._sts = None
        self._iam_client = None
        self._iam = None
        keychain_enabled = defaults.extras.get(DISABLE_KEYRING) is not True
        vault = FiggyVault(keychain_enabled=keychain_enabled)
        self._sts_cache: CacheManager = CacheManager(file_override=STS_SESSION_CACHE_PATH, vault=vault)
        self._role_name_prefix = os.getenv(FIGGY_ROLE_PREFIX_OVERRIDE_ENV, FIGGY_ROLE_NAME_PREFIX)

    def __get_iam_user(self):
        self._defaults.user = self.__get_iam_resource().CurrentUser().user_name
        return self._defaults.user

    def __get_iam_resource(self):
        if not self._iam:
            self._iam = self.__bastion_session.resource('iam')

        return self._iam

    def __get_iam_client(self):
        if not self._iam_client:
            self._iam_client = self.__bastion_session.client('iam')

        return self._iam_client

    def __get_ssm(self):
        if not self._ssm:
            self._ssm = SsmDao(self.__bastion_session.client('ssm'))
        return self._ssm

    def __get_sts(self):
        if not self._sts:
            self._sts = self.__bastion_session.client('sts')
        return self._sts

    def get_mfa_serial(self) -> Optional[str]:
        response = self.__get_iam_client().list_mfa_devices(UserName=self._defaults.user)
        devices = response.get('MFADevices', [])
        log.info(f'Found MFA devices: {devices}.')
        return devices[0].get('SerialNumber') if devices else None

    def get_session(self, assumable_role: AssumableRole, prompt: bool, exit_on_fail=True) -> boto3.Session:
        forced = False
        log.info(f"Getting session for role: {assumable_role.role_arn} in env: {assumable_role.run_env.env}")
        attempts = 0
        while True:
            try:
                if prompt and not forced:
                    forced = True
                    raise InvalidSessionError("Forcing new session due to prompt.")

                creds: FiggyAWSSession = self._sts_cache.get_val(assumable_role.role.full_name)

                if creds:
                    session = boto3.Session(
                        aws_access_key_id=creds.access_key,
                        aws_secret_access_key=creds.secret_key,
                        aws_session_token=creds.token,
                        region_name=self._defaults.region
                    )

                    if not self._is_valid_session(session):
                        self._utils.validate(attempts < self._MAX_ATTEMPTS,
                                             f"Failed to authenticate with AWS after {attempts} attempts. Exiting. ")

                        attempts = attempts + 1
                        log.info("Invalid session detected in cache. Raising session error.")
                        raise InvalidSessionError("Invalid Session Detected")

                    log.info("Valid session returned from cache.")
                    return session
                else:
                    raise InvalidSessionError("Forcing new session, cache is empty.")
            except (FileNotFoundError, NoCredentialsError, InvalidSessionError) as e:
                try:
                    if self._defaults.mfa_enabled:
                        self._defaults.mfa_serial = self.get_mfa_serial()
                        color = Utils.default_colors() if self._defaults.colors_enabled else None
                        mfa = SecretsManager.generate_mfa(self._defaults.user) if self._defaults.auto_mfa else \
                                                            Input.get_mfa(display_hint=True, color=color)

                        response = self.__get_sts().assume_role(RoleArn=assumable_role.role_arn,
                                                                RoleSessionName=Utils.sanitize_session_name(
                                                                    self._defaults.user),
                                                                DurationSeconds=self._defaults.session_duration,
                                                                SerialNumber=self._defaults.mfa_serial,
                                                                TokenCode=mfa)
                    else:
                        response = self.__get_sts().assume_role(RoleArn=assumable_role.role_arn,
                                                                RoleSessionName=Utils.sanitize_session_name(
                                                                    self._defaults.user),
                                                                DurationSeconds=self._defaults.session_duration)

                    session = FiggyAWSSession.from_sts_response(response)
                    log.info(f"Got session response: {response}")
                    self._sts_cache.write(assumable_role.role.full_name, session)
                except (ClientError, ParamValidationError) as e:
                    if isinstance(e, ParamValidationError) or "AccessDenied" == e.response['Error']['Code']:
                        if exit_on_fail:
                            self._utils.error_exit(f"Error authenticating with AWS from Bastion Profile:"
                                                   f" {self._defaults.provider_config.profile_name}: {e}")
                    else:
                        if exit_on_fail:
                            log.error(f"Failed to authenticate due to error: {e}")
                            self._utils.error_exit(
                                f"Error getting session for role: {assumable_role.role_arn} "
                                f"-- Are you sure you have permissions?")

                    raise e

    def get_assumable_roles(self):
        if self.is_role_session():
            user_roles = [self._defaults.role.role]
        else:
            ROLE_PATH = f'/figgy/users/{self.__get_iam_user()}/roles'
            user_roles = self.__get_ssm().get_parameter(ROLE_PATH)
            self._utils.stc_validate(user_roles is not None and user_roles != "[]",
                                     "Something is wrong with your user's configuration with Figgy. "
                                     "Unable to find any eligible roles for your user. Please contact your"
                                     " administrator.")

            user_roles = json.loads(user_roles)

        environments = self.__get_ssm().get_all_parameters([PS_FIGGY_ACCOUNTS_PREFIX], option='OneLevel')
        names: List[str] = [env.get('Name') for env in environments]
        parameters = self.__get_ssm().get_parameter_values(names)
        assumable_roles: List[AssumableRole] = []

        for param in parameters:
            env_name = param.get('Name').split('/')[-1]
            account_id = param.get('Value')

            for role in user_roles:
                assumable_roles.append(AssumableRole(
                    run_env=RunEnv(env=env_name, account_id=account_id),
                    role=Role(role, full_name=f'{FIGGY_ROLE_NAME_PREFIX}{env_name}-{role}'),
                    account_id=account_id,
                    provider_name=Provider.AWS_BASTION.value,
                    profile=None
                ))

        return assumable_roles

    def is_role_session(self):
        """
        For sandbox demos, where users aren't coming from user accounts, we want to skip looking up user -> role.
        :return: bool - Is this session originating from a role?
        """
        creds = self.__bastion_session.get_credentials().get_frozen_credentials()

        return hasattr(creds, 'token') and creds.token is not None

    def cleanup_session_cache(self):
        self._sts_cache.wipe_cache()



import os
from json import JSONDecodeError
from typing import List

import boto3
import logging
import base64
import json
from abc import ABC, abstractmethod

from botocore.exceptions import NoCredentialsError, ParamValidationError, ClientError

from figcli.config import *
from figcli.models.assumable_role import AssumableRole
from figcli.models.aws_session import FiggyAWSSession
from figcli.models.defaults.defaults import CLIDefaults
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.auth.provider.session_provider import SessionProvider
from figcli.svcs.vault import FiggyVault
from figcli.utils.secrets_manager import SecretsManager
from figcli.utils.utils import InvalidSessionError, Utils

log = logging.getLogger(__name__)


class SSOSessionProvider(SessionProvider, ABC):
    _MAX_ATTEMPTS = 5

    def __init__(self, defaults: CLIDefaults):
        super().__init__(defaults)
        self._utils = Utils(defaults.colors_enabled)
        self._sts = boto3.client('sts')
        keychain_enabled = defaults.extras.get(DISABLE_KEYRING) is not True
        vault = FiggyVault(keychain_enabled=keychain_enabled)
        self._sts_cache: CacheManager = CacheManager(file_override=STS_SESSION_CACHE_PATH, vault=vault)
        self._saml_cache: CacheManager = CacheManager(file_override=SAML_SESSION_CACHE_PATH, vault=vault)

    @abstractmethod
    def get_assumable_roles(self) -> List[AssumableRole]:
        pass

    @abstractmethod
    def cleanup_session_cache(self):
        pass

    @abstractmethod
    def get_saml_assertion(self, prompt: bool = False):
        pass

    def get_session(self, assumable_role: AssumableRole, prompt: bool, exit_on_fail=True) -> boto3.Session:
        """
        Creates a session in the specified ENV for the target role from a SAML assertion returned by SSO authentication.
        Args:
            assumable_role: AssumableRole - The role to be leveraged to authenticate this session
            prompt: If prompt is set, we will not use a cached session and will generate new sessions for okta and mgmt.
            exit_on_fail: Exit the program if this session hydration fails.

        returns: Hydrated session for role + account that match the specified one in the provided AssumableRole
        """

        role_arn = f"arn:aws:iam::{assumable_role.account_id}:role/{assumable_role.role.full_name}"
        principal_arn = f"arn:aws:iam::{assumable_role.account_id}:saml-provider/{assumable_role.provider_name}"
        forced = False
        log.info(f"Getting session for role: {role_arn} in env: {assumable_role.run_env.env} "
                 f"with principal: {principal_arn}")
        attempts = 0
        while True:
            try:
                if prompt and not forced:
                    forced = True
                    raise InvalidSessionError("Forcing new session due to prompt.")

                creds: FiggyAWSSession = self._sts_cache.get_val(assumable_role.role.full_name)
                if creds:
                    session = boto3.Session(
                        aws_access_key_id=creds.access_key,
                        aws_secret_access_key=creds.secret_key,
                        aws_session_token=creds.token,
                        region_name=self._defaults.region
                    )

                    if not self._is_valid_session(session):
                        self._utils.validate(attempts < self._MAX_ATTEMPTS,
                                             f"Failed to authenticate with AWS after {attempts} attempts. Exiting. ")

                        attempts = attempts + 1
                        log.info("Invalid session detected in cache. Raising session error.")
                        raise InvalidSessionError("Invalid Session Detected")

                    log.info("Valid session returned from cache.")
                    return session
                else:
                    raise InvalidSessionError("Forcing new session, cache is empty.")
            except (FileNotFoundError, JSONDecodeError, NoCredentialsError, InvalidSessionError) as e:
                try:
                    # Todo Remove requiring raw saml and instead work with b64 encoded saml?
                    try:
                        assertion: str = self._saml_cache.get_val_or_refresh(SAML_ASSERTION_CACHE_KEY,
                                                                             self.get_saml_assertion, prompt,
                                                                             max_age=SAML_ASSERTION_MAX_AGE)
                        encoded_assertion = base64.b64encode(assertion.encode('utf-8')).decode('utf-8')
                        response = self._sts.assume_role_with_saml(RoleArn=role_arn,
                                                                   PrincipalArn=principal_arn,
                                                                   SAMLAssertion=encoded_assertion)
                    except ClientError:
                        log.info("Refreshing SAML assertion, auth failed with cached or refreshed version.")
                        assertion = self.get_saml_assertion(prompt)
                        encoded_assertion = base64.b64encode(assertion.encode('utf-8')).decode('utf-8')
                        response = self._sts.assume_role_with_saml(RoleArn=role_arn,
                                                                   PrincipalArn=principal_arn,
                                                                   SAMLAssertion=encoded_assertion)

                    response['Credentials']['Expiration'] = "cleared"
                    session = FiggyAWSSession.from_sts_response(response)
                    self._saml_cache.write(SAML_ASSERTION_CACHE_KEY, assertion)
                    log.info(f"Got session response: {response}")
                    self._sts_cache.write(assumable_role.role.full_name, session)
                except (ClientError, ParamValidationError) as e:
                    if isinstance(e, ParamValidationError) or "AccessDenied" == e.response['Error']['Code']:
                        if exit_on_fail:
                            self._utils.error_exit(f"Error authenticating with AWS from SAML Assertion: {e}")
                    else:
                        if exit_on_fail:
                            print(e)
                            self._utils.error_exit(
                                f"Error getting session for role: {role_arn} -- Are you sure you have permissions?")

                    raise e



import logging
from dataclasses import dataclass
from typing import Dict

import requests

log = logging.getLogger(__name__)


@dataclass
class CircleResponse:
    status_code: int
    data: Dict


class CircleCi:

    def __init__(self, token: str):
        self._token = token

    def submit_job(self, repository: str, tag: str = None, revision: str = None,
                   extra_params: Dict = None) -> CircleResponse:
        """
        Submit a new job to CircleCi
        :param repository: circleci repo to trigger build for
        :param tag: tag to trigger a build on. Cannot be supplied if revision is set
        :param revision: githash to trigger a build for. Cannot be supplied if tag is set
        :param extra_params: Dict - Map of environment variables to add to the CircleCi build
        :return: CircleCi response. Response code and data in the CircleCi API response.
        """
        if tag and revision:
            raise ValueError("You cannot provide both a `tag` and a `revision`, you must select one or the other.")

        request_url = f"https://circleci.com/api/v1.1/project/github/Snagajob/{repository}?circle-token={self._token}"
        headers, content = {'Content-Type': 'application/json'}, {}

        if tag:
            content['tag'] = tag
        elif revision:
            content['revision'] = revision

        if extra_params:
            content['build_parameters'] = extra_params

        result = requests.post(request_url, json=content, headers=headers)
        return CircleResponse(data=result.json(), status_code=result.status_code)



import curses
from concurrent.futures import ThreadPoolExecutor, as_completed

import npyscreen
from botocore.exceptions import ClientError
from npyscreen import MLTreeMultiSelect, NPSTreeData, NPSApp, Form, TreeLineSelectable

from figcli.commands.config.delete import Delete
from figcli.commands.config.get import Get
from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.config import ConfigDao
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *
from figcli.utils.utils import Utils
from figcli.views.rbac_limited_config import RBACLimitedConfigView

log = logging.getLogger(__name__)


class Browse(ConfigCommand, npyscreen.NPSApp):

    def __init__(self, ssm_init: SsmDao, colors_enabled: bool, context: ConfigContext,
                 get_command: Get, delete_command: Delete, config_view: RBACLimitedConfigView):
        super().__init__(browse, colors_enabled, context)
        self._ssm = ssm_init
        self._get = get_command
        self._config_view = config_view
        self.selected_ps_paths = []
        self.deleted_ps_paths = []
        self.dirs = set()
        self._utils = Utils(colors_enabled)
        self._delete = delete_command
        self.prefix = context.prefix

    def add_children(self, prefix: str, td_node: NPSTreeData, all_children: set = None,
                     first_children: set = None):
        """
        Looks up all children at `prefix` and recursively adds them to the TreeData node.
        Adds all children to the TreeData node passed in, and recurses down the tree of children.

        Once the original total lookup of all PS nodes is done, all_children and first_children are progressively
        filtered and passed down with smaller data sets as the tree is built.
        Args:
            prefix: Prefix to query and built a tree from. I.E. /app
            td_node: The node to append found first children too
            all_children: List of ALL children of the passed in prefix's node.
            first_children:  List of the FIRST children of the passed in prefix's node.
                            (Immediate children, not grandchildren)
        """
        with ThreadPoolExecutor(max_workers=10) as pool:
            first_res = None
            all_res = None
            if first_children is None:
                first_res = pool.submit(self._config_view.get_config_names, prefix=prefix, one_level=True)

            if all_children is None:
                all_res = pool.submit(self._config_view.get_config_names, prefix=prefix)

            if first_res is not None:
                first_children = set(first_res.result())

            if all_res is not None:
                all_children = set(all_res.result())

        diff_children = all_children.difference(first_children)

        child_dirs = set(list(map(lambda x: f"{prefix}/{x.replace(f'{prefix}/', '', 1).split('/')[0]}", diff_children)))
        self.dirs = self.dirs | child_dirs

        for child in sorted(first_children):
            child_name = child.replace(prefix, '', 1)
            if child_name != '/':
                td_node.newChild(content=child_name, selectable=True, expanded=False)

        for child_dir in sorted(child_dirs):
            child_name = child_dir.replace(prefix, '', 1)
            if child_name != '/':
                grand_children = set(list(filter(lambda x: x.startswith(f'{prefix}{child_name}/'), all_children)))
                calculated_first_children = set(list(filter(
                    lambda x: len(x.split('/')) == len(prefix.split('/')) + 2, grand_children)))

                dir_node = td_node.newChild(content=child_dir.replace(prefix, '', 1), selectable=False, expanded=False)
                self.add_children(child_dir, dir_node, all_children=grand_children,
                                  first_children=calculated_first_children)

    @AnonymousUsageTracker.track_command_usage
    def _browse(self):
        browse_app = BrowseApp(self, self._config_view)
        browse_app.run()

    def _print_val(self, path, val, desc):
        if val is not None:
            print(f"\r\n{self.c.fg_bl}Key:{self.c.rs} {path}")
            print(f"{self.c.fg_bl}Value:{self.c.rs} {val}\r\n")
            desc = desc if desc else DESC_MISSING_TEXT
            print(f"{self.c.fg_bl}Description:{self.c.rs} {desc}\r\n")

    def _safe_delete(self, ps_name: str):
        try:
            self._ssm.delete_parameter(ps_name)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ParameterNotFound':
                return

        print(f"{self.c.fg_gr}Deleted: {ps_name}{self.c.rs}")

    def _perform_deletions(self) -> None:
        deleted = []  # type: List[str]
        for path in self.deleted_ps_paths:
            if path in self.dirs:
                all_children = set(list(map(lambda x: x['Name'],
                                            self._ssm.get_all_parameters([path], option='Recursive'))))
                delete_children = Input.y_n_input(f"You have selected a DIRECTORY to delete: {path}. "
                      f"Do you want to delete ALL children of: {path}?", default_yes=False)

                if delete_children:
                    for child in all_children:
                        if self._delete.delete_param(child):
                            deleted.append(child)
            else:
                if path not in deleted:
                    delete_it = Input.y_n_input(f"Delete {path}? ", default_yes=True)

                    if delete_it:
                        if self._delete.delete_param(path):
                            deleted.append(path)

    def _print_values(self) -> None:
        for path in self.selected_ps_paths:
            if path in self.dirs:
                notify, selection = False, ''
                print(f"\r\nIt appears you selected a directory: {path}.")
                while not self._utils.is_valid_selection(selection, notify):
                    selection = input(f"Would you like to print values for all children of {path}? (Y/n): ")
                    selection = selection if selection != '' else 'y'
                if selection.lower() == 'y':
                    all_children = set(list(map(lambda x: x['Name'],
                                                self._ssm.get_all_parameters([path], option='Recursive'))))
                    for child in all_children:
                        val, desc = self._get.get_val_and_desc(child)
                        self._print_val(child, val, desc)
            else:
                val, desc = self._get.get_val_and_desc(path)
                self._print_val(path, val, desc)

    @AnonymousUsageTracker.track_command_usage
    @VersionTracker.notify_user
    def execute(self):
        self._browse()
        self._perform_deletions()
        self._print_values()


class DeletableNPSTreeData(NPSTreeData):
    """
    Extends NPSTreeData class to add deleted property, so we may mark individual nodes as deleted.
    """

    def __init__(self, content=None, parent=None, selected=False, selectable=True,
                 highlight=False, expanded=True, ignoreRoot=True, sort_function=None, deleted=False):
        super().__init__(content=content, parent=parent, selected=selected, selectable=selectable,
                         highlight=highlight, expanded=expanded, ignoreRoot=ignoreRoot,
                         sort_function=sort_function)
        self.deleted = deleted


class BrowseApp(NPSApp):
    """
    Extends NPSApp class to add support for getting a list of deleted tree nodes that were marked for deletion.
    """

    def __init__(self, browse: Browse, config_view: RBACLimitedConfigView):
        self._browse = browse
        self._config_view = config_view

    def main(self):
        global npy_form
        npy_form = Form(

            name="Browse Parameters: - 'e' to expand, 'c' to collapse, <s> to select, <d> to delete, "
                 "<Tab> & <Shift+Tab> moves cursor between `OK` and `Tree` views.", )

        tree = npy_form.add(LogicalMLTree)

        td = DeletableNPSTreeData(content='Root', selectable=True, expanded=True, ignoreRoot=True)
        start = Utils.millis_since_epoch()
        children = []
        if self._browse.prefix:
            prefix_child = td.newChild(content=self._browse.prefix, selectable=False, expanded=False)
            children = [prefix_child]
        else:
            log.info(f"--{Utils.get_first(prefix)} missing, defaulting to normal browse tree.")

            for namespace in self._config_view.get_authorized_namespaces():
                child = td.newChild(content=namespace, selectable=False, expanded=False)
                children.append(child)

        for child in children:
            self._browse.dirs.add(child.getContent())

        futures = []

        with ThreadPoolExecutor(max_workers=10) as pool:
            for child in children:
                futures.append(pool.submit(self._browse.add_children, child.getContent(), child))

        for future in as_completed(futures):
            pass

        tree.values = td
        npy_form.edit()
        selection_objs = tree.get_selected_objects(return_node=True)

        for selection in selection_objs:
            full_path = ''
            while selection._parent is not None:
                full_path = selection.content + full_path
                selection = selection._parent
            self._browse.selected_ps_paths.append(full_path)

        delete_objs = tree.get_objects_to_delete(return_node=True)
        for selection in delete_objs:
            full_path = ''
            while selection._parent is not None:
                full_path = selection.content + full_path
                selection = selection._parent
            self._browse.deleted_ps_paths.append(full_path)


class TreeLineMultiFunctionSelect(TreeLineSelectable):
    """
    Extends functionality of TreeLineSelectable to provide DELETE toggling
    """
    CAN_DELETE_DELETED = '[DELETE]'
    CANNOT_SELECT_DELETED = '[DELETE_DIR]'

    def _print_select_controls(self):
        SELECT_DISPLAY = None

        if self._tree_real_value.selectable:
            if self.value.deleted:
                SELECT_DISPLAY = self.CAN_DELETE_DELETED
            elif self.value.selected:
                SELECT_DISPLAY = self.CAN_SELECT_SELECTED
            else:
                SELECT_DISPLAY = self.CAN_SELECT
        else:
            if self.value.deleted:
                SELECT_DISPLAY = self.CANNOT_SELECT_DELETED
            elif self.value.selected:
                SELECT_DISPLAY = self.CANNOT_SELECT_SELECTED
            else:
                SELECT_DISPLAY = self.CANNOT_SELECT

        if self.do_colors():
            attribute_list = self.parent.theme_manager.findPair(self, 'CONTROL')
        else:
            attribute_list = curses.A_NORMAL

        # python2 compatibility
        if isinstance(SELECT_DISPLAY, bytes):
            SELECT_DISPLAY = SELECT_DISPLAY.decode()

        self.add_line(self.rely,
                      self.left_margin + self.relx,
                      SELECT_DISPLAY,
                      self.make_attributes_list(SELECT_DISPLAY, attribute_list),
                      self.width - self.left_margin,
                      )

        return len(SELECT_DISPLAY)


class LogicalMLTree(MLTreeMultiSelect):
    _contained_widgets = TreeLineMultiFunctionSelect

    """
    Sets ups Logical instead of insane key bindings for ML Tree Viewer.

    Also disables select_cascades by default.
    
    Also provides ability to set `delete` on nodes, and query all deleted nodes
    """

    def __init__(self, screen, select_cascades=False, *args, **keywords):
        super(MLTreeMultiSelect, self).__init__(screen, select_cascades=select_cascades, *args, **keywords)
        self.select_cascades = select_cascades

    def set_up_handlers(self):
        super(MLTreeMultiSelect, self).set_up_handlers()
        self.handlers.update({
            ord('c'): self.h_collapse_tree,
            ord('e'): self.h_expand_tree,
            ord('C'): self.h_collapse_all,
            ord('E'): self.h_expand_all,
            ord('d'): self.h_delete,
            ord('s'): self.h_select,
            ord('r'): self.h_refresh
        })

    def h_select(self, ch):
        vl = self.values[self.cursor_line]
        vl_to_set = not vl.selected
        if self.select_cascades:
            for v in self._walk_tree(vl, only_expanded=False, ignore_root=False):
                if v.selectable:
                    v.selected = vl_to_set
        else:
            vl.selected = vl_to_set
        if self.select_exit:
            self.editing = False
            self.how_exited = True
        self.display()

    def h_refresh(self, ch):
        npyscreen.blank_terminal()
        npy_form.edit()
        # npy_form.useable_space

    def h_delete(self, ch):
        vl = self.values[self.cursor_line]
        vl_to_set = not vl.deleted
        if self.select_cascades:
            for v in self._walk_tree(vl, only_expanded=False, ignore_root=False):
                if v.selectable:
                    v.deleted = vl_to_set
        else:
            vl.deleted = vl_to_set
        if self.select_exit:
            self.editing = False
            self.how_exited = True
        self.display()

    def get_objects_to_delete(self, return_node=True):
        for v in self._walk_tree(self._myFullValues, only_expanded=False, ignore_root=False):
            if v.deleted:
                if return_node:
                    yield v
                else:
                    yield self._get_content(v)



from figcli.config.commands import *
from figcli.config.aws import *
from figcli.config.style import *
from figcli.config.constants import *
from figcli.config.communication import *
from figcli.config.sso import *
from figcli.config.figgy import *


from botocore.exceptions import ClientError
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.config.style.style import FIGGY_STYLE
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *
from figcli.commands.types.config import ConfigCommand
from figcli.commands.config_context import ConfigContext
from figcli.data.dao.config import ConfigDao
from figcli.models.replication_config import ReplicationConfig
from figcli.data.dao.ssm import SsmDao


class Delete(ConfigCommand):

    def __init__(self, ssm_init: SsmDao, config_init: ConfigDao, context: ConfigContext, colors_enabled: bool,
                 config_completer: WordCompleter):
        super().__init__(delete, colors_enabled, context)
        self._ssm = ssm_init
        self._config = config_init
        self._utils = Utils(colors_enabled)
        self._config_completer = config_completer

        # Prompts for this file
        self._del_message = [
            ('class:', 'PS Name to Delete: ')
        ]

    def delete_param(self, key) -> bool:
        """
        Manages safe deletion through the CLI. Prevents deletion of replication sources. Prompts user for deletion of
        replication destinations.
        Args:
            key: PS Name / Key

        Returns: bool - T/F based on whether a parameter was actually deleted.
        """
        sources = self._config.get_cfgs_by_src(key, self.run_env)  # type: List[ReplicationConfig]
        repl_conf = self._config.get_config_repl(key, self.run_env)  # type: ReplicationConfig

        if len(sources) > 0:
            print(f"{self.c.fg_rd}You're attempting to delete a key that is the source for at least one "
                  f"replication config.{self.c.rs}\n{self.c.fg_bl}{key}{self.c.rs} is actively replicating to these"
                  f" destinations:\n")
            for src in sources:
                print(f"Dest: {self.c.fg_rd}{src.destination}{self.c.rs}. "
                      f"This config was created by {self.c.fg_bl}{src.user}{self.c.rs}. ")

            print(
                f"\r\n{self.c.fg_bl}{key}{self.c.rs} is a replication SOURCE. Deleting this source would effectively BREAK "
                f"replication to the above printed destinations. You may NOT delete sources that are actively "
                f"replicating. Please delete the above printed DESTINATIONS first. "
                f"Once they have been deleted, you will be allowed to delete this "
                f"SOURCE.")
            return False
        elif repl_conf is not None:
            selection = "unselected"
            while selection.lower() != "y" and selection.lower() != "n":
                repl_msg = [
                    (f'class:{self.c.rd}', f"{key} is an active replication destination created by "),
                    (f'class:{self.c.bl}', f"{repl_conf.user}. "),
                    (f'class:{self.c.rd}', f"Do you want to ALSO delete this replication config and "
                                           f"permanently delete {key}? "),
                    (f'class:', "(y/N): ")]
                selection = prompt(repl_msg, completer=WordCompleter(['Y', 'N']), style=FIGGY_STYLE)
                selection = selection if selection != '' else 'n'
                if selection.strip().lower() == "y":
                    self._config.delete_config(key, self.run_env)
                    self._ssm.delete_parameter(key)
                    print(f"{self.c.fg_gr}{key} and replication config destination deleted successfully.{self.c.rs}")
                    return True
                elif selection.strip().lower() == "n":
                    return False

        else:
            delete_msg = f"{self.c.fg_gr}{key} deleted successfully.{self.c.rs}\r\n"
            try:
                print(f"Trying delete {key}")
                self._ssm.delete_parameter(key)
            except ClientError as e:
                if e.response['Error']['Code'] == 'ParameterNotFound':
                    print(delete_msg)
                    return True
                elif "AccessDeniedException" == e.response['Error']['Code']:
                    print(f"{self.c.fg_rd}You do not have permissions to delete: {key}{self.c.rs}")
                    return False
                else:
                    raise

            print(delete_msg)
            return True

    def _delete_param(self):
        """
        Prompts user for a parameter name to delete, then deletes
        """
        # Add all keys
        key, notify, delete_another = None, False, True

        while not self._utils.is_valid_input(key, f"PS Name", notify) or delete_another:
            key = prompt(self._del_message, style=FIGGY_STYLE,
                         completer=self._config_completer)
            notify = True
            try:
                if self._utils.is_valid_input(key, 'PS Parameter', False):
                    if self.delete_param(key):
                        if key in self._config_completer.words:
                            self._config_completer.words.remove(key)
                else:
                    continue
            except ClientError as e:
                error_code = e.response['Error']['Code']
                if "AccessDeniedException" == error_code:
                    print(f"\n\nYou do not have permissions to a new config value at the path: "
                          f"{self.c.fg_bl} {key} {self.c.rs}")
                    print(f"Developers may add keys under the following namespaces: "
                          f"{self.c.fg_bl} {self.c.rs}")
                    print(f"Error message: {e.response['Error']['Message']}")
                elif "ParameterNotFound" == error_code:
                    print(f"{self.c.fg_rd} The specified Name: {self.c.rs} {self.c.fg_bl} {key}{self.c.rs}"
                          f"{self.c.fg_rd} does not exist in the selected environment. Please try again.', red {self.c.rs}")
                else:
                    print(f"Exception caught attempting to delete config: {e.response['Message']}")

            print()
            to_continue = input(f"Delete another? (Y/n): ")
            to_continue = to_continue if to_continue != '' else 'y'
            delete_another = to_continue.lower() == "y"


    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._delete_param()




# OKTA Constants
OKTA_SESSION_DURATION = 43200
SUPPORTED_OKTA_FACTOR_TYPES = ['GOOGLE', 'OKTA']

FIGGY_KEYRING_NAMESPACE = "figgy"
FIGGY_KEYRING_ASSERTION_NAME = "okta_assertion"

SAML_ASSERTION_CACHE_KEY = 'assertion'
SAML_ASSERTION_MAX_AGE = 60 * 5 * 1000  # Assertions don't last long, 5 mins max reuse



import base64

from figcli.utils.utils import *


class KmsSvc:
    """
    Provides access to KMS apis
    """

    def __init__(self, boto_kms_client):
        self._kms = boto_kms_client

    def is_encrypted(self, obj: str):
        try:
            self.decrypt(obj)
            return True
        except Exception as e:
            return False

    def decrypt(self, base64_ciphertext):
        ciphertext = base64.b64decode(base64_ciphertext)
        return self._kms.decrypt(CiphertextBlob=ciphertext)[u"Plaintext"].decode()

    def decrypt_with_context(self, base64_ciphertext, context: Dict):
        ciphertext = base64.b64decode(base64_ciphertext)
        return self._kms.decrypt(
            CiphertextBlob=ciphertext,
            EncryptionContext=context)[u"Plaintext"].decode()



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.dev.get import DevGet
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *


class DevDelete(FiggyTest):
    def __init__(self, extra_args=""):
        print(f"Testing `figgy config {Utils.get_first(delete)} --env {DEFAULT_ENV}`")
        super().__init__(
            pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(delete)} --env {DEFAULT_ENV} '
                          f'{extra_args} --skip-upgrade',
                          timeout=5, encoding='utf-8'), extra_args=extra_args)
        self._child.delayafterread = .01
        self._child.delaybeforesend = .5

    def run(self):
        print(f"Testing DELETE for {param_1}")
        self.delete(param_1)

    def delete(self, name, delete_another=False):
        self.expect('.*PS Name to Delete.*')
        self.sendline(name)
        print(f"Delete sent for {name}")
        self.expect(f'.*deleted successfully.*Delete another.*')

        print("Validating delete success.")
        get = DevGet(extra_args=self.extra_args)
        get.get(name, DELETE_ME_VALUE, expect_missing=True)

        if delete_another:
            self.sendline('y')
        else:
            self.sendline('n')

        print("Successful delete validated.")



from os import getcwd
from typing import Tuple

from prompt_toolkit import prompt

from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import *

log = logging.getLogger(__name__)


class Generate(ConfigCommand):

    def __init__(self, colors_enabled: bool, config_context: ConfigContext):
        super().__init__(generate, colors_enabled, config_context)
        self._from_path = config_context.ci_config_path if config_context.ci_config_path else Utils.find_figgy_json()
        self._utils = Utils(colors_enabled)
        self._errors_detected = False
        self.example = f"{self.c.fg_bl}{CLI_NAME} config {self.command_printable} " \
            f"--env dev --config /path/to/config{self.c.rs}"

    @staticmethod
    def _get_service_name_and_version(service_name: str) -> Tuple[str, int]:
        base_matcher = re.compile(r"^([A-Za-z0-9_-]+)([0-9]+)$")
        result = base_matcher.match(service_name)
        base_name = result.group(1) if result else service_name
        version = int(result.group(2)) if result else 1

        return base_name, version

    def _generate(self):
        from_config = self._utils.get_ci_config(self._from_path)
        service_name = self._utils.get_namespace(from_config).split('/')[2]
        current_ns = self._utils.get_namespace(from_config)

        base_name, version = self._get_service_name_and_version(service_name)
        base_name = base_name if not base_name.endswith('-') else base_name[:-1]
        new_service_name = f'{base_name}-{version + 1}'

        new_name = prompt(f'Please select a new service name, it CANNOT be: {service_name}:  ',
                          default=new_service_name)
        self._utils.validate(new_name != service_name, f"You must select a new service name that differs from the one"
        f"designated in your source figgy.json file. "
        f"(NOT {service_name})")
        new_ns = f'{self.context.defaults.service_ns}/{new_name}/'

        # Update all configs destinations to leverage new namespace. Easiest to search/replace across everything.
        output_string = json.dumps(from_config)
        output_string = output_string.replace(current_ns[:-1], new_ns[:-1])
        new_config = json.loads(output_string)

        # Remove existing configs that will be replicated
        new_config[CONFIG_KEY] = []

        # Configure replicate_from block
        new_config[REPL_FROM_KEY] = {
            SOURCE_NS_KEY: from_config.get(REPL_FROM_KEY, {}).get(SOURCE_NS_KEY, current_ns),
            PARAMETERS_KEY: from_config.get(REPL_FROM_KEY, {}).get(PARAMETERS_KEY, [])
        }

        for param in from_config.get(CONFIG_KEY, []):
            new_config[REPL_FROM_KEY][PARAMETERS_KEY].append(self._utils.get_parameter_only(param))

        formatted_config = self._utils.format_config(new_config)
        current_dir = getcwd()
        output_file = prompt(f'Write new config here?: ', default=f'{current_dir}/{new_name}-config.json')
        self._utils.is_valid_input(output_file, "output_file", True)

        with open(output_file, "w") as file:
            file.write(json.dumps(formatted_config, sort_keys=False, indent=4))

        print(f'{self.c.fg_gr}New config successfully generated at location: {output_file}{self.c.rs}')

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self._generate()



import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.config import *
from figcli.utils.utils import *
from subprocess import check_output
from figcli.test.cli.dev.configure import DevConfigure
from figcli.test.cli.data.configure import DataConfigure
from figcli.test.cli.devops.configure import DevOpsConfigure
import time

class CLIInit:

    @staticmethod
    def init(role: str):
        assert role in user_types
        if role == usr_dev:
            DevConfigure().run()
        elif role == usr_data:
            DataConfigure().run()
        elif role == usr_devops:
            DevOpsConfigure().run()

        figgy = pexpect.spawn(f'{CLI_NAME} config {Utils.get_first(get)} --env {DEFAULT_ENV} --prompt --skip-upgrade',
                                    timeout=5, encoding='utf-8')
        figgy.expect('.*Please input the MFA.*')
        mfa = check_output(["mfa", "otp", os.environ[MFA_USER_ENV_KEY]])
        mfa = str(mfa, 'utf-8').rstrip()
        print(f"Sending mfa {mfa}")
        figgy.sendline(mfa)
        figgy.expect('.*What type.*')
        figgy.sendline(role)
        result = figgy.expect(['.*input a PS Name.*', '.*Invalid MFA code.*'])
        count = 0
        # If mfa is invalid, keep trying.
        while result == 1 and count < 15:
            count = count + 1
            mfa = check_output(["mfa", "otp", os.environ[MFA_USER_ENV_KEY]])
            mfa = str(mfa, 'utf-8').rstrip()
            print(f"Retrying MFA: {mfa} - Attempt # {count}")
            figgy.sendline(mfa)
            result = figgy.expect(['.*input a PS Name.*', '.*Invalid MFA code.*'])
            time.sleep(10)

        time.sleep(15)





import json
import logging
from typing import List

from prompt_toolkit.completion import WordCompleter

from figcli.config import *
from figcli.data.dao.ssm import SsmDao
from figcli.models.role import Role
from figcli.models.run_env import RunEnv
from figcli.svcs.cache_manager import CacheManager
from figcli.svcs.config import ConfigService
from figcli.utils.utils import Utils

log = logging.getLogger(__name__)

class RBACLimitedConfigView:
    """
    Returns limited sets of configuration names based on the user's role's access
    """

    def __init__(self, role: Role, cache_mgr: CacheManager, ssm: SsmDao, config_svc: ConfigService):
        self._role = role
        self._cache_mgr = cache_mgr
        self._config_svc = config_svc
        self._ssm = ssm
        self.rbac_role_ns_path = f'{figgy_ns}/rbac/{self._role.role}/namespaces'
        self.rbac_role_kms_path = f'{figgy_ns}/rbac/{self._role.role}/keys'
        self._config_completer = None

    def get_authorized_namespaces(self) -> List[str]:
        """
        Looks up the user-defined namespaces that users of this type can access. This enables us to prevent the
        auto-complete from showing parameters the user doesn't actually have access to.

        If the user's defined accessible namespaces are not available, then returns all root namespaces, regardless
        of whether use is authorized to perform GET requests against those values.

        Leverages an expiring local cache to save ~200ms on each figgy bootstrap
        """
        cache_key = f'{self._role.role}-authed-nses'

        es, authed_nses = self._cache_mgr.get_or_refresh(cache_key, self._ssm.get_parameter, self.rbac_role_ns_path)

        if authed_nses:
            if isinstance(authed_nses, str):
                authed_nses = json.loads(authed_nses)
        else:
            es, authed_nses = self._cache_mgr.get_or_refresh(cache_key, self._config_svc.get_root_namespaces)

        if not isinstance(authed_nses, list):
            raise ValueError(f"Invalid value found at path: {self.rbac_role_ns_path}. It must be a valid json List[str]")

        return authed_nses

    def get_authorized_kms_keys(self) -> List[str]:
        """
        Looks up the user-defined KMS keys that users of this type can access. This enables us to prevent the
        auto-complete from showing keys the user doesn't actually have access to.

        Leverages an expiring local cache to save ~200ms on each figgy bootstrap
        """
        cache_key = f'{self._role.role}-authed-keys'

        es, authed_nses = self._cache_mgr.get_or_refresh(cache_key, self._ssm.get_parameter, self.rbac_role_kms_path)

        if authed_nses:
            authed_nses = json.loads(authed_nses)

        if not isinstance(authed_nses, list):
            raise ValueError(f"Invalid value found at path: {self.rbac_role_kms_path}. It must be a valid json List[str]")

        return authed_nses

    def get_authorized_key_id(self, authorized_key_name: str, run_env: RunEnv) -> str:
        """
        Returns the appropriate KMS Key ID for a provided authorized key name.
        :param authorized_key_name: KMS Key the provider user has access to.
        :param run_env: Run environment associated with this key
        :return: KMS Key id of the associated key.
        """
        cache_key = f'kms-{authorized_key_name}-{run_env.env}'
        key_path = f'/figgy/kms/{authorized_key_name}-key-id'

        if authorized_key_name not in self.get_authorized_kms_keys():
            raise ValueError(f"You do not have access to encrypt with the KMS key: {authorized_key_name}")

        es, key_id = self._cache_mgr.get_or_refresh(cache_key, self._ssm.get_parameter, key_path)
        return key_id

    @Utils.trace
    def get_config_names(self, prefix: str = None, one_level: bool = False) -> List[str]:
        all_names = sorted(self._config_svc.get_parameter_names())
        authed_nses = self.get_authorized_namespaces()
        new_names = [] if authed_nses or prefix else all_names

        if prefix:
            is_child = False
            for ns in authed_nses:
                if prefix.startswith(ns):
                    is_child = True

            if not is_child and authed_nses:
                raise ValueError(f"Provided prefix of {prefix} is not in a valid authorized namespace.")

            authed_nses = [prefix]

        for ns in authed_nses:
            filtered_names = [name for name in all_names if name.startswith(ns)]
            new_names = new_names + filtered_names

        if one_level:
            if prefix:
                new_names = [name for name in new_names if len(name.split('/')) == len(prefix) + 1]
            else:
                new_names = authed_nses

        return new_names

    @Utils.trace
    def get_config_completer(self):
        """
        This is used to be a slow operation since it involves pulling all parameter names from Parameter Store.
        It's best to be lazy loaded only if the dependent command requires it. It's still best to be lazy loaded,
        but it is much faster now that we have implemented caching of existing parameter names in DynamoDb and
        locally.
        """
        # Not the most efficient, but plenty fast since we know the # of authed_nses is gonna be ~<=5
        # Tested at 30k params and it takes ~25ms
        if not self._config_completer:
            self._config_completer = WordCompleter(self.get_config_names(), sentence=True, match_middle=True)

        # print(f"Cache Count: {len(all_names)}")
        return self._config_completer



import logging
import re

from figcli.config import *
from botocore.exceptions import ClientError
from npyscreen import Form, MultiLineEdit, NPSApp, BoxTitle
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from figcli.commands.config_context import ConfigContext
from figcli.commands.types.config import ConfigCommand
from figcli.data.dao.ssm import SsmDao
from figcli.input import Input
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker
from figcli.utils.utils import Utils
from figcli.views.rbac_limited_config import RBACLimitedConfigView

log = logging.getLogger(__name__)


class MultiLineInput(BoxTitle):
    _contained_widget = MultiLineEdit


class EditApp(NPSApp):
    def __init__(self, name: str, value: str, description: str):
        self.name = name
        self.value = value if value else ''
        self.description = description if description else ''
        self.value_box = None
        self.description_box = None

    def main(self):
        edit_form = Form()
        self.value_box = edit_form.add(MultiLineInput, value=self.value, name=f'Input value for: {self.name}',
                                       max_height=12)
        self.description_box = edit_form.add(MultiLineInput, value=self.description, name="Optional Description: ",
                                             max_height=12)
        print(f"Value box: {self.value_box} - {self.description_box}")
        edit_form.edit()


class Edit(ConfigCommand):

    def __init__(self, ssm_init: SsmDao, colors_enabled: bool, config_context: ConfigContext,
                 config_view: RBACLimitedConfigView, config_completer: WordCompleter):
        super().__init__(edit, colors_enabled, config_context)
        self._ssm = ssm_init
        self._config_view = config_view
        self._utils = Utils(colors_enabled)
        self._config_completer = config_completer

    def edit(self) -> None:
        """
        Allows a user to define a PS name and add or edit a parameter at that location. Uses NPYscreen editor.
        """
        key = Input.input('Please input a PS Name: ', completer=self._config_completer)

        try:
            value, desc = self._ssm.get_parameter_with_description(key)
            edit_app = EditApp(key, value, desc)
            edit_app.run()

            value, desc = edit_app.value_box.value, edit_app.description_box.value
            log.info(f"Edited value: {value} - description: {desc}")

            is_secret = Input.is_secret()
            parameter_type, kms_id = SSM_SECURE_STRING if is_secret else SSM_STRING, None
            if is_secret:
                valid_keys = self._config_view.get_authorized_kms_keys()
                if len(valid_keys) > 1:
                    key_name = Input.select_kms_key(valid_keys)
                else:
                    key_name = valid_keys[0]

                kms_id = self._config_view.get_authorized_key_id(key_name, self.run_env)

            if not self._utils.is_valid_input(key, f"Parameter name", True) \
                    or not self._utils.is_valid_input(value, key, True):
                self._utils.error_exit("Invalid input detected, please resolve the issue and retry.")

            self._ssm.set_parameter(key, value, desc, parameter_type, key_id=kms_id)
            print(f"{self.c.fg_gr}{key} saved successfully.{self.c.rs}")
        except ClientError as e:
            if "AccessDeniedException" == e.response['Error']['Code']:
                denied = "AccessDeniedException" == e.response['Error']['Code']
                if denied and "AWSKMS; Status Code: 400;" in e.response['Error']['Message']:
                    print(f"\n{self.c.fg_rd}You do not have access to decrypt the value of: {key}{self.c.rs}")
                elif denied:
                    print(f"\n{self.c.fg_rd}You do not have access to Parameter: {key}{self.c.rs}")
                else:
                    raise
            else:
                self._utils.error_exit(f"{self.c.fg_rd}Exception caught attempting to add config: {e}{self.c.rs}")

    @VersionTracker.notify_user
    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self.edit()



from figcli.commands.help_context import HelpContext
from figcli.commands.types.help import HelpCommand
from figcli.config import *
from figcli.input import Input
from figcli.svcs.observability.anonymous_usage_tracker import AnonymousUsageTracker
from figcli.svcs.observability.version_tracker import VersionTracker, FiggyVersionDetails
from figcli.svcs.upgrade_manager import UpgradeManager
from figcli.utils.utils import Utils


class Upgrade(HelpCommand):
    """
    Drives the --version command
    """

    def __init__(self, help_context: HelpContext):
        super().__init__(version, help_context.defaults.colors_enabled, help_context)
        self.tracker = VersionTracker(self.context.defaults)
        self.upgrade_mgr = UpgradeManager(help_context.defaults.colors_enabled)
        self._utils = Utils(colors_enabled=help_context.defaults.colors_enabled)

    def upgrade(self):
        latest_version: FiggyVersionDetails = self.tracker.get_version()

        if latest_version.version == VERSION:
            print(f'{self.c.fg_bl}You are currently using the latest version of {CLI_NAME}: {self.c.rs}'
                  f'{self.c.fg_gr}{VERSION}{self.c.rs}')

        elif self.tracker.upgrade_available(VERSION, latest_version.version):
            print(f'{self.c.fg_yl}------------------------------------------{self.c.rs}')
            print(f' New version: {self.c.rs}{self.c.fg_gr}{latest_version.version}{self.c.rs} is more '
                  f'recent than your version: {self.c.fg_gr}{VERSION}{self.c.rs}')
            print(f'{self.c.fg_yl}------------------------------------------{self.c.rs}')

            if self._utils.is_mac():
                print(f"\nMacOS auto-upgrade is supported. Performing auto-upgrade.")
                self.install_mac(latest_version)
            print(f"{self.c.fg_gr}Installation successful! Exiting. Rerun `{CLI_NAME}` "
                  f"to use the latest version!{self.c.rs}")
        else:
            print(f'{self.c.fg_yl}------------------------------------------{self.c.rs}')
            print(f'Your version: {self.c.rs}{self.c.fg_gr}{latest_version.version}{self.c.rs} is more '
                  f'recent than the current recommended version of {CLI_NAME}: {self.c.fg_gr}{VERSION}{self.c.rs}')
            print(f'{self.c.fg_yl}------------------------------------------{self.c.rs}')

            if self._utils.is_mac():
                selection = Input.y_n_input(f'Would you like to revert to the current recommended version '
                                            f'of {CLI_NAME}?')
                if selection:
                    self.install_mac(latest_version)

    def install_mac(self, latest_version: FiggyVersionDetails):
        install_path = self.upgrade_mgr.get_install_path()
        self.upgrade_mgr.install_mac_onedir(install_path, latest_version.version)

    @AnonymousUsageTracker.track_command_usage
    def execute(self):
        self.upgrade()



from pathlib import Path
from figcli.config.aws import *
from figcli.config.constants import *

# Root subcommand types
version = frozenset({'version'})
command = frozenset({'command'})
resource = frozenset({'resource'})
configure = frozenset({'configure'})

# Resource types
config = frozenset({'config'})
iam = frozenset({'iam'})
login = frozenset({'login'})

resources = config | iam

# Config Sub Command definitions
sync = frozenset({'sync'})
put = frozenset({'put'})
restore = frozenset({'restore'})
point_in_time = frozenset({'point-in-time'})
delete = frozenset({'delete'})
cleanup = frozenset({'cleanup'})
get = frozenset({'get'})
edit = frozenset({'edit'})
list_com = frozenset({'list'})
share = frozenset({'share'})
promote = frozenset({'promote'})
ci_path = frozenset({'config'})
info = frozenset({'info'})
browse = frozenset({'browse'})
prompt_com = frozenset({'prompt'})
audit = frozenset({'audit'})
dump = frozenset({'dump'})
replication_only = frozenset({'replication-only'})
manual = frozenset({'manual'})
env = frozenset({'env'})
prefix = frozenset({'prefix'})
out = frozenset({'out'})
skip_upgrade = frozenset({'skip-upgrade'})
service = frozenset({'service'})
debug = frozenset({'debug'})
copy_from = frozenset({'copy-from'})
generate = frozenset({'generate'})
from_path = frozenset({'from'})
validate = frozenset({'validate'})
profile = frozenset({'profile'})

# IAM sub commands
export = frozenset({'export'})
all_profiles = frozenset({'all-profiles'})
role = frozenset({'role'})

# argparse options
help = frozenset({'help'})
required = frozenset({'required'})
action = frozenset({'action'})
store_true = 'store_true'

# help commands
sandbox = frozenset({'sandbox'})
upgrade = frozenset({'upgrade'})

# Maps CLI `--options` for each argument, and sets flags if necessary
arg_options = {
    config: {
        cleanup: {
            config: {action: None, required: False},
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        delete: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        get: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        list_com: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        put: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        edit: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        restore: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            point_in_time: {action: store_true, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        share: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        sync: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            replication_only: {action: store_true, required: False},
            config: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            copy_from: {action: None, required: False},
            profile: {action: None, required: False},
        },
        browse: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            prefix: {action: None, required: False},
            profile: {action: None, required: False},
        },
        dump: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            prefix: {action: None, required: False},
            out: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        audit: {
            info: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        promote: {
            info: {action: store_true, required: False},
            env: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        generate: {
            info: {action: store_true, required: False},
            env: {action: None, required: False},
            from_path: {action: None, required: False},
            role: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        validate: {
            info: {action: store_true, required: False},
            prompt_com: {action: store_true, required: False},
            env: {action: None, required: False},
            config: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
    },
    iam: {
        export: {
            info: {action: store_true, required: False},
            env: {action: None, required: False},
            skip_upgrade: {action: store_true, required: False},
            all_profiles: {action: store_true, required: False},
            role: {action: None, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
    },
    login: {
        login: {
            info: {action: store_true, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
            profile: {action: None, required: False},
        },
        sandbox: {
            info: {action: store_true, required: False},
            skip_upgrade: {action: store_true, required: False},
            debug: {action: store_true, required: False},
        }
    }
}

# Merge key suffixes
merge_uri_suffix = ":uri"
empty_uri_suffix = ''
merge_suffixes = [merge_uri_suffix, empty_uri_suffix]

# Supported commands by resource
config_commands = [sync, put, edit, delete, cleanup, get, share, generate,
                   list_com, browse, audit, dump, restore, promote, validate]
iam_commands = [export]
help_commands = [configure, version, login, sandbox, upgrade]

# Used to build out parser, map of resource to sub-commands
resource_map = {
    config: config_commands,
    iam: iam_commands,
    login: [login, sandbox]
}

options = ci_path | info


# KMS Key Types / Mapping
kms_app = 'app'
kms_data = 'data'
kms_devops = 'devops'
kms_keys = [kms_app, kms_data, kms_devops]

# Validation Supported types
plugin = "plugin"
cve = "cve"

# Command to option requirement map
REQ_OPTION_MAP = {
    cleanup: [ci_path],
    delete: [],
    get: [],
    list_com: [],
    put: [],
    restore: [],
    share: [],
    sync: [ci_path],
    edit: [],
    generate: [ci_path]
}



from typing import Optional, List, Set

from figcli.commands.command_context import CommandContext
from figcli.models.defaults.defaults import CLIDefaults
from figcli.models.run_env import RunEnv


class HelpContext(CommandContext):
    """
    Contextual information for HelpCommands, including _what_ command resources were passed in. Help commands
    often don't have standard "resource" or "command" blocks, instead they may ONLY have --optional parameters
    """
    def __init__(self, resource: Optional[frozenset], command: Optional[frozenset],
                 options: Optional[Set[frozenset]], run_env: Optional[RunEnv], defaults: Optional[CLIDefaults]):
        super().__init__(run_env, resource, defaults=defaults)

        self.resource = resource
        self.command = command
        self.options = options



from prompt_toolkit.styles import Style

# Prompt Colors
from figcli.config.style.color import Color
from figcli.config.style.terminal_factory import TerminalFactory

colors: Color = TerminalFactory(True).instance().get_colors()

FIGGY_STYLE = Style.from_dict({
    colors.bl: colors.bl_val,
    colors.gr: colors.gr_val,
    colors.rd: colors.rd_val,
    colors.yl: colors.yl_val
})


import pexpect
from figcli.test.cli.config import *
from figcli.test.cli.figgy import FiggyTest
from figcli.utils.utils import *
import os


class ConfigureOkta(FiggyTest):
    def __init__(self, role_type: str):
        self._role_type = role_type
        super().__init__(pexpect.spawn(f'{CLI_NAME} --{Utils.get_first(configure)} --skip-upgrade',
                                       encoding='utf-8', timeout=10))
        # self._child = pexpect.spawn(f'{CLI_NAME} --{Utils.get_first(configure)} --skip-upgrade',
        #                                encoding='utf-8', timeout=10)

    def run(self):
        self.step(f"Testing `{CLI_NAME} --{Utils.get_first(configure)}`")
        user_name = os.environ.get(OKTA_SSO_USER)
        password = os.environ.get(OKTA_SSO_PASSWORD)
        embed_url = os.environ.get(OKTA_EMBED_URL)
        mfa_secret = os.environ.get(OKTA_MFA_SECRET)

        self._child.expect('.*select.*OKTA.*')
        self._child.sendline('OKTA')
        self._child.expect('.*OKTA username.*')
        self._child.sendline(user_name)
        self._child.expect('.*OKTA password.*')
        self._child.sendline(password)
        self._child.expect('.*mfa_enabled.*')
        self._child.sendline('n')
        self._child.expect('.*Use Multi-factor.*')
        self._child.sendline('y')
        self._child.expect('.*auto_mfa.*')
        self._child.sendline('n')
        self._child.expect('.*Would you.*generate.*')
        self._child.sendline('y')
        self._child.expect('.*auto-generate.*')
        self._child.sendline(mfa_secret)
        self._child.expect('.*app_link.*')
        self._child.sendline('n')
        self._child.expect('.*Embed Link.*')
        self._child.sendline(embed_url)
        self._child.expect('.*GOOGLE.*')
        self._child.sendline('y')
        self._child.expect('.*Options are.*')
        self._child.sendline(self._role_type)
        self._child.expect('.*Options.*dev.*')
        self._child.sendline('dev')
        self._child.expect('.*region.*')
        self._child.sendline('y')
        self._child.expect('.*colors.*')
        self._child.sendline('n')
        self._child.expect('.*weirdness.*')
        self._child.sendline('n')
        self._child.expect('.*anonymous_metrics_enabled.*')
        self._child.sendline('n')
        self._child.expect('.*anonymou.*')
        self._child.sendline('n')
        self._child.expect('.*overwrite.*')
        self._child.expect('.*Setup successful.*')
        self.step(f"OKTA with role: {self._role_type} configured successfully")
